# pragma: no cover
from os import environ

from koheesio.__about__ import __version__, _about
from koheesio.context import Context
from koheesio.logger import LoggingFactory
from koheesio.models import BaseModel, ExtraParamsMixin
from koheesio.steps import Step, StepOutput
from koheesio.utils import convert_str_to_bool

_koheesio_print_logo = convert_str_to_bool(environ.get("KOHEESIO__PRINT_LOGO", "True"))
_logo_printed = False
ABOUT = _about()
VERSION = __version__

__all__ = [
    "ABOUT",
    "BaseModel",
    "Context",
    "ExtraParamsMixin",
    "LoggingFactory",
    "Step",
    "StepOutput",
    "VERSION",
]


def print_logo() -> None:
    global _logo_printed
    global _koheesio_print_logo

    if not _logo_printed and _koheesio_print_logo:
        print(ABOUT)
        _logo_printed = True


print_logo()



from pathlib import Path
from unittest.mock import MagicMock

from koheesio.spark.writers import BatchOutputMode
from koheesio.spark.writers.file_writer import FileFormat, FileWriter


def test_path_validator():
    output_path = Path("/test/output_path")
    file_writer = FileWriter(output_mode=BatchOutputMode.APPEND, format=FileFormat.parquet, path=output_path)
    assert isinstance(file_writer.path, str)


def test_execute(dummy_df, mocker):
    path = "expected_path"
    output_mode = BatchOutputMode.APPEND
    options = {"option1": "value1", "option2": "value2"}
    format = FileFormat.parquet
    writer = FileWriter(df=dummy_df, output_mode=output_mode, path=path, format=format, **options)

    mock_df_writer = MagicMock()

    from koheesio.spark.utils.connect import is_remote_session

    if is_remote_session():
        from pyspark.sql import DataFrame as SparkDataFrame
        from pyspark.sql.connect.dataframe import DataFrame as ConnectDataFrame

        mocker.patch.object(SparkDataFrame, "write", mock_df_writer)
        mocker.patch.object(ConnectDataFrame, "write", mock_df_writer)
    else:
        from pyspark.sql import DataFrame

        mocker.patch.object(DataFrame, "write", mock_df_writer)

    mock_df_writer.options.return_value = mock_df_writer

    writer.execute()
    mock_df_writer.options.assert_called_with(**options)
    mock_df_writer.save.assert_called_with(path=path, format=format.value, mode=output_mode.value)



"""SQL Transform module

SQL Transform module provides an easy interface to transform a dataframe using SQL. This SQL can originate from a
string or a file and may contain placeholders for templating.
"""

from koheesio.models.sql import SqlBaseStep
from koheesio.spark.transformations import Transformation
from koheesio.spark.utils import SPARK_MINOR_VERSION, on_databricks
from koheesio.utils import get_random_string


class SqlTransform(SqlBaseStep, Transformation):
    """
    SQL Transform module provides an easy interface to transform a dataframe using SQL.

    This SQL can originate from a string or a file and may contain placeholder (parameters) for templating.

    - Placeholders are identified with `${placeholder}`.
    - Placeholders can be passed as explicit params (params) or as implicit params (kwargs).

    Example sql script:

    ```sql
    SELECT id, id + 1 AS incremented_id, ${dynamic_column} AS extra_column
    FROM ${table_name}
    ```
    """

    def execute(self) -> Transformation.Output:
        table_name = get_random_string(prefix="sql_transform")
        self.params = {**self.params, "table_name": table_name}

        from koheesio.spark.utils.connect import is_remote_session

        if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session() and self.df.isStreaming and not on_databricks():
            raise RuntimeError(
                "SQL Transform is not supported in remote sessions with streaming dataframes."
                "See https://issues.apache.org/jira/browse/SPARK-45957"
                "It is fixed in PySpark 4.0.0"
            )

        self.df.createOrReplaceTempView(table_name)
        query = self.query

        self.output.df = self.spark.sql(query)



import datetime

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.concat import Concat
from koheesio.spark.utils import show_string

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)


string_data = [[1, "text1", "text2", "text3"], [2, "text4", "text5", "text6"]]
string_schema = ["id", "_1", "_2", "_3"]


@pytest.mark.parametrize(
    "input_values,input_data,input_schema,expected",
    [
        (
            # description: Test Concat without Spacer
            # input values
            dict(
                columns=["_1", "_2", "_3"],
                target_column="result",
            ),
            # input data and schema
            string_data,
            string_schema,
            # expected data
            ["text1text2text3", "text4text5text6"],
        ),
        (
            # description: Test Concat with Spacer
            # input values
            dict(
                columns=["_1", "_2", "_3"],
                spacer="|",
                target_column="result",
            ),
            # input data and schema
            string_data,
            string_schema,
            # expected data
            ["text1|text2|text3", "text4|text5|text6"],
        ),
        (
            # description: Test Concat with integer type
            # input values
            dict(
                columns=["_1", "id"],
                target_column="result",
            ),
            # input data and schema
            string_data,
            string_schema,
            # expected data
            ["text11", "text42"],
        ),
        (
            # description: Test that Concat retains column order
            # input values
            dict(
                columns=["_3", "_1"],
                target_column="result",
            ),
            # input data and schema
            string_data,
            string_schema,
            # expected data
            ["text3text1", "text6text4"],
        ),
        (
            # description: Test Concat with a float type
            # input values
            dict(
                columns=["float_col", "_1"],
                target_column="result",
            ),
            # input data and schema
            [[1.0, "text1"], [2.0, "text2"]],
            ["float_col", "_1"],
            # expected data
            ["1.0text1", "2.0text2"],
        ),
        (
            # description: Test Concat with a null field
            # input values
            dict(
                columns=["none_col", "_1"],
                target_column="result",
            ),
            # input data and schema
            [["text1", "text2"], [None, "text2"]],
            ["none_col", "_1"],
            # expected data
            ["text1text2", None],
        ),
        (
            # description: Test Concat with a datetime field
            # input values
            dict(
                columns=["datetime_col", "_1"],
                target_column="result",
            ),
            # input data and schema
            [[datetime.datetime(1997, 2, 28), "text"]],
            ["datetime_col", "_1"],
            # expected data
            ["1997-02-28 00:00:00text"],
        ),
        (
            # description: Test Concat with a array field, including a null value
            # input values
            dict(
                columns=["array_col_1", "array_col_2"],
                target_column="result",
            ),
            # input data and schema
            [[["text1", "text2"], ["text3", None]]],
            ["array_col_1", "array_col_2"],
            # expected data
            [["text1", "text2", "text3", None]],
        ),
        (
            # description: Test Concat with a array field and spacer
            # input values
            dict(columns=["array_col_1", "array_col_2"], target_column="result", spacer="--"),
            # input data and schema
            [[["text1", "text2"], ["text3", "text4"]]],
            ["array_col_1", "array_col_2"],
            # expected data
            ["text1--text2--text3--text4"],
        ),
        (
            # description: Test Concat with a array field, with null value and spacer
            # This is the same as the example in the class docstring
            # input values
            dict(columns=["array_col_1", "array_col_2"], target_column="result", spacer="--"),
            # input data and schema
            [[["text1", "text2"], ["text3", None]]],
            ["array_col_1", "array_col_2"],
            # expected data
            ["text1--text2--text3"],
        ),
        (
            # description: Test Concat with a numeric field
            # input values
            dict(
                columns=["numeric_col", "_1"],
                target_column="result",
            ),
            # input data and schema
            [[1, "text1"], [2, "text2"]],
            ["numeric_col", "_1"],
            # expected data
            ["1text1", "2text2"],
        ),
    ],
)
def test_happy_flow(input_values, input_data, input_schema, expected, spark):
    input_df = spark.createDataFrame(input_data, input_schema)

    concat = Concat(**input_values)
    target_column = concat.target_column
    output_df = concat.transform(input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{show_string(output_df, 20, 20, False)}")

    actual = [row[target_column] for row in output_df.select(target_column).collect()]
    assert actual == expected


# TODO: not supported atm
#
#     def test_concatenate_nested_base_1(self):
#         df = (
#             Concatenate(
#                 df=self.sample_nested_df,
#                 params={"source_columns": ["payload.col2.col3", "payload.col2.col4"]},
#                 target_alias="payload.col2.col5",
#                 columns_to_flatten=["payload{}.col2{}.col3", "payload{}.col2{}.col4"],
#             )
#             .execute()
#             .df
#         )
#         result = df.rdd.map(lambda r: r.payload).collect()
#         self.assertEqual(result[0]["col2"]["col5"], "value1value2")



"""Repartition Transformation"""

from typing import Optional

from koheesio.models import Field, ListOfColumns, model_validator
from koheesio.spark.transformations import ColumnsTransformation


class Repartition(ColumnsTransformation):
    """
    Wrapper around DataFrame.repartition

    With repartition, the number of partitions can be given as an optional value. If this is not provided, a default
    value is used. The default number of partitions is defined by the spark config 'spark.sql.shuffle.partitions', for
    which the default value is 200 and will never exceed the number of rows in the DataFrame (whichever is value is
    lower).

    If columns are omitted, the entire DataFrame is repartitioned without considering the particular values in the
    columns.

    Parameters
    ----------
    columns : Optional[Union[str, List[str]]], optional, default=None
        Name of the source column(s). If omitted, the entire DataFrame is repartitioned without considering the
        particular values in the columns. Alias: column
    num_partitions : Optional[int], optional, default=None
        The number of partitions to repartition to. If omitted, the default number of partitions is used as defined by
        the spark config 'spark.sql.shuffle.partitions'.

    Example
    -------
    ```python
    Repartition(
        column=["c1", "c2"], num_partitions=3
    )  # results in 3 partitions
    Repartition(column="c1", num_partitions=2)  # results in 2 partitions
    Repartition(column=["c1", "c2"])  # results in <= 200 partitions
    Repartition(num_partitions=5)  # results in 5 partitions
    ```
    """

    columns: Optional[ListOfColumns] = Field(default="", alias="column", description="Name of the source column(s)")
    num_partitions: Optional[int] = Field(
        default=None,
        alias="numPartitions",
        description="The number of partitions to repartition to. If omitted, the default number of partitions is used "
        "as defined by the spark config 'spark.sql.shuffle.partitions'.",
    )

    @model_validator(mode="before")
    def _validate_field_and_num_partitions(cls, values: dict) -> dict:
        """Ensure that at least one of the fields 'columns' and 'num_partitions' is provided."""
        columns_value = values.get("columns") or values.get("column")
        num_partitions_value = values.get("numPartitions") or values.get("num_partitions")

        if not columns_value and not num_partitions_value:
            raise ValueError("The fields 'columns' and 'num_partitions' cannot both be empty!")

        values["numPartitions"] = num_partitions_value
        return values

    def execute(self) -> ColumnsTransformation.Output:
        # Prepare columns input:
        columns = self.df.columns if self.columns == ["*"] else self.columns
        # Prepare repartition input:
        #  num_partitions comes first, but if it is not provided it should not be included as None.
        repartition_inputs = [i for i in [self.num_partitions, *columns] if i]  # type: ignore
        self.output.df = self.df.repartition(*repartition_inputs)



"""
This module contains async implementation of HTTP step.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union
import asyncio
import warnings

from aiohttp import BaseConnector, ClientSession, TCPConnector
from aiohttp_retry import ExponentialRetry, RetryClient, RetryOptionsBase
import nest_asyncio  # type: ignore[import-untyped]
import yarl

from pydantic import Field, SecretStr, field_validator, model_validator

from koheesio.asyncio import AsyncStep, AsyncStepOutput
from koheesio.models import ExtraParamsMixin
from koheesio.steps.http import HttpMethod


# noinspection PyUnresolvedReferences
class AsyncHttpStep(AsyncStep, ExtraParamsMixin):
    """
    Asynchronous HTTP step for making HTTP requests using aiohttp.

    Parameters
    ----------
    client_session : Optional[ClientSession]
        Aiohttp ClientSession.
    url : List[yarl.URL]
        List of yarl.URL.
    retry_options : Optional[RetryOptionsBase]
        Retry options for the request.
    connector : Optional[BaseConnector]
        Connector for the aiohttp request.
    headers : Optional[Dict[str, Union[str, SecretStr]]]
        Request headers.

    Output
    ------
    responses_urls : Optional[List[Tuple[Dict[str, Any], yarl.URL]]]
        List of responses from the API and request URL.

    Examples
    --------
    ```python
    import asyncio
    from aiohttp import ClientSession
    from aiohttp.connector import TCPConnector
    from aiohttp_retry import ExponentialRetry
    from koheesio.asyncio.http import AsyncHttpStep
    from yarl import URL
    from typing import Dict, Any, Union, List, Tuple


    # Initialize the AsyncHttpStep
    async def main():
        session = ClientSession()
        urls = [
            URL("https://example.com/api/1"),
            URL("https://example.com/api/2"),
        ]
        retry_options = ExponentialRetry()
        connector = TCPConnector(limit=10)
        headers = {"Content-Type": "application/json"}
        step = AsyncHttpStep(
            client_session=session,
            url=urls,
            retry_options=retry_options,
            connector=connector,
            headers=headers,
        )

        # Execute the step
        responses_urls = await step.get()

        return responses_urls


    # Run the main function
    responses_urls = asyncio.run(main())
    ```
    """

    client_session: Optional[ClientSession] = Field(default=None, description="Aiohttp ClientSession", exclude=True)
    url: List[yarl.URL] = Field(
        default_factory=list,
        alias="urls",
        description="""Expecting list, as there is no value in executing async request for one value.
        yarl.URL is preferable, because params/data can be injected into URL instance""",
        exclude=True,
    )
    retry_options: Optional[RetryOptionsBase] = Field(
        default=None, description="Retry options for the request", exclude=True
    )
    connector: Optional[BaseConnector] = Field(
        default=None, description="Connector for the aiohttp request", exclude=True
    )
    headers: Dict[str, Union[str, SecretStr]] = Field(
        default_factory=dict,
        description="Request headers",
        alias="header",
        exclude=True,
    )

    timeout: None = Field(default=None, description="[Optional] Request timeout")

    method: Union[str, HttpMethod] = Field(
        default=HttpMethod.GET,
        description="What type of Http call to perform. One of 'get', 'post', 'put', 'delete'. Defaults to 'get'.",
    )

    class Output(AsyncStepOutput):
        """Output class for Step"""

        responses_urls: Optional[List[Tuple[Dict[str, Any], yarl.URL]]] = Field(
            default=None, description="List of responses from the API and request URL", repr=False
        )

    def __tasks_generator(self, method: HttpMethod) -> List[asyncio.Task]:
        """
        Generate a list of tasks for making HTTP requests.

        Parameters
        ----------
        method : HttpMethod
            The HTTP method to use for the requests.

        Returns
        -------
        List[Task]
            A list of asyncio tasks for making HTTP requests.
        """
        tasks = [
            asyncio.create_task(
                self.request(
                    method=method,
                    url=u,
                    headers=self.get_headers(),
                )
            )
            for u in self.url
        ]

        return tasks

    @model_validator(mode="after")
    def _move_extra_params_to_params(self) -> AsyncHttpStep:
        """
        Move extra_params to params dict.

        Returns
        -------
        AsyncHttpStep
            The updated AsyncHttpStep instance.
        """
        return self

    async def _execute(self, tasks: List[asyncio.Task]) -> List[Tuple[Dict[str, Any], yarl.URL]]:
        """
        Execute the HTTP requests asynchronously.

        Parameters
        ----------
        tasks : List[Task]
            A list of asyncio tasks for making HTTP requests.

        Returns
        -------
        List[Tuple[Dict[str, Any], yarl.URL]]
            A list of response data and corresponding request URLs.
        """
        self._init_session()
        try:
            responses_urls = await asyncio.gather(*tasks)
        finally:
            if self.client_session:
                await self.client_session.close()
            await self.__retry_client.close()

        return responses_urls

    def _init_session(self) -> None:
        """
        Initialize the aiohttp session and retry client.
        """
        self.connector = self.connector or TCPConnector(limit=10)
        self.client_session = self.client_session or ClientSession(connector=self.connector)
        self.retry_options = self.retry_options or ExponentialRetry()
        # Disable pylint warning: attribute is not initialized in __init__
        # pylint: disable=W0201
        self.__retry_client = RetryClient(
            client_session=self.client_session, logger=self.log, retry_options=self.retry_options, raise_for_status=True
        )

    @field_validator("timeout")
    def validate_timeout(cls, timeout: Any) -> None:
        """
        Validate the 'timeout' field.

        Parameters
        ----------
        timeout : Any
            The value of the 'timeout' field.

        Raises
        ------
        ValueError
            If 'data' is not allowed in AsyncHttpStep.
        """
        if timeout:
            raise ValueError("timeout is not allowed in AsyncHttpStep. Provide timeout through retry_options.")

    def get_headers(self) -> Union[None, dict]:
        """
        Get the request headers.

        Returns
        -------
        Optional[Dict[str, Union[str, SecretStr]]]
            The request headers.
        """
        _headers = None

        if self.headers:
            _headers = {k: v.get_secret_value() if isinstance(v, SecretStr) else v for k, v in self.headers.items()}

            for k, v in self.headers.items():
                if isinstance(v, SecretStr):
                    self.headers[k] = v.get_secret_value()

        return _headers or self.headers

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def set_outputs(self, response) -> None:  # type: ignore[no-untyped-def]
        """
        Set the outputs of the step.

        Parameters
        ----------
        response : Any
            The response data.
        """
        warnings.warn("set outputs is not implemented in AsyncHttpStep.")

    # noinspection PyMethodMayBeStatic
    def get_options(self) -> None:
        """
        Get the options of the step.
        """
        warnings.warn("get_options is not implemented in AsyncHttpStep.")

    # Disable pylint warning: method was expected to be 'non-async'
    # pylint: disable=W0236
    async def request(  # type: ignore[no-untyped-def]
        self,
        method: HttpMethod,
        url: yarl.URL,
        **kwargs,
    ) -> Tuple[Dict[str, Any], yarl.URL]:
        """
        Make an HTTP request.

        Parameters
        ----------
        method : HttpMethod
            The HTTP method to use for the request.
        url : yarl.URL
            The URL to make the request to.
        kwargs : Any
            Additional keyword arguments to pass to the request.

        Returns
        -------
        Tuple[Dict[str, Any], yarl.URL]
            A tuple containing the response data and the request URL.
        """
        async with self.__retry_client.request(method=method, url=url, **kwargs) as response:
            res = await response.json()

        return res, response.request_info.url

    # Disable pylint warning: method was expected to be 'non-async'
    # pylint: disable=W0236
    # noinspection PyMethodOverriding
    async def get(self) -> List[Tuple[Dict[str, Any], yarl.URL]]:
        """
        Make GET requests.

        Returns
        -------
        List[Tuple[Dict[str, Any], yarl.URL]]
            A list of response data and corresponding request URLs.
        """
        tasks = self.__tasks_generator(method=HttpMethod.GET)
        responses_urls = await self._execute(tasks=tasks)

        return responses_urls

    # Disable pylint warning: method was expected to be 'non-async'
    # pylint: disable=W0236
    async def post(self) -> List[Tuple[Dict[str, Any], yarl.URL]]:
        """
        Make POST requests.

        Returns
        -------
        List[Tuple[Dict[str, Any], yarl.URL]]
            A list of response data and corresponding request URLs.
        """
        tasks = self.__tasks_generator(method=HttpMethod.POST)
        responses_urls = await self._execute(tasks=tasks)

        return responses_urls

    # Disable pylint warning: method was expected to be 'non-async'
    # pylint: disable=W0236
    async def put(self) -> List[Tuple[Dict[str, Any], yarl.URL]]:
        """
        Make PUT requests.

        Returns
        -------
        List[Tuple[Dict[str, Any], yarl.URL]]
            A list of response data and corresponding request URLs.
        """
        tasks = self.__tasks_generator(method=HttpMethod.PUT)
        responses_urls = await self._execute(tasks=tasks)

        return responses_urls

    # Disable pylint warning: method was expected to be 'non-async'
    # pylint: disable=W0236
    async def delete(self) -> List[Tuple[Dict[str, Any], yarl.URL]]:
        """
        Make DELETE requests.

        Returns
        -------
        List[Tuple[Dict[str, Any], yarl.URL]]
            A list of response data and corresponding request URLs.
        """
        tasks = self.__tasks_generator(method=HttpMethod.DELETE)
        responses_urls = await self._execute(tasks=tasks)

        return responses_urls

    def execute(self) -> None:
        """
        Execute the step.

        Raises
        ------
        ValueError
            If the specified HTTP method is not implemented in AsyncHttpStep.
        """
        # By design asyncio does not allow its event loop to be nested. This presents a practical problem:
        #   When in an environment where the event loop is already running
        #   it’s impossible to run tasks and wait for the result.
        #   Trying to do so will give the error “RuntimeError: This event loop is already running”.
        #   The issue pops up in various environments, such as web servers, GUI applications and in
        #   Jupyter/DataBricks notebooks.
        nest_asyncio.apply()

        map_method_func = {
            HttpMethod.GET: self.get,
            HttpMethod.POST: self.post,
            HttpMethod.PUT: self.put,
            HttpMethod.DELETE: self.delete,
        }

        if self.method not in map_method_func:
            raise ValueError(f"Method {self.method} not implemented in AsyncHttpStep.")

        self.output.responses_urls = asyncio.run(map_method_func[self.method]())  # type: ignore[index]


class AsyncHttpGetStep(AsyncHttpStep):
    """
    Represents an asynchronous HTTP GET step.

    This class inherits from the AsyncHttpStep class and specifies the HTTP method as GET.

    Attributes:
        method (HttpMethod): The HTTP method for the step, set to HttpMethod.GET.

    """

    method: HttpMethod = HttpMethod.GET



import pytest

import pyspark.sql.types as T

from koheesio.spark import AnalysisException
from koheesio.spark.readers.file_loader import (
    AvroReader,
    CsvReader,
    FileFormat,
    FileLoader,
    JsonReader,
    OrcReader,
    ParquetReader,
)

pytestmark = pytest.mark.spark


@pytest.fixture()
def json_file(data_path):
    return f"{data_path}/readers/json_file/dummy_simple.json"


@pytest.fixture()
def csv_comma_file(data_path):
    return f"{data_path}/readers/csv_file/dummy.csv"


@pytest.fixture()
def csv_semicolon_file(data_path):
    return f"{data_path}/readers/csv_file/dummy_semicolon.csv"


@pytest.fixture()
def parquet_file(data_path):
    return f"{data_path}/readers/delta_file"


@pytest.fixture()
def avro_file(data_path):
    return f"{data_path}/readers/avro_file"


@pytest.fixture()
def orc_file(data_path):
    return f"{data_path}/readers/orc_file"


def test_file_loader(csv_comma_file):
    # test schema
    expected_data = [
        {"value": "string,int,float"},
        {"value": "string1,1,1.0"},
        {"value": "string2,2,2.0"},
        {"value": "string3,3,3.0"},
    ]
    schema = "value STRING"
    reader = FileLoader(path=csv_comma_file, header=True, schema=schema, lineSep="\n")
    assert reader.schema_ == schema
    df = reader.read()
    actual_data = [row.asDict() for row in df.collect()]
    print(f"{actual_data = }")
    assert actual_data == expected_data


def test_invalid_file_format(csv_comma_file):
    with pytest.raises(ValueError):
        FileLoader(format="invalid_format", path=csv_comma_file, header=True)


def test_csv_reader(csv_comma_file, csv_semicolon_file):
    expected_data = [
        {"string": "string1", "int": 1, "float": 1.0},
        {"string": "string2", "int": 2, "float": 2.0},
        {"string": "string3", "int": 3, "float": 3.0},
    ]

    schema = "string STRING, int INT, float FLOAT"

    # comma separated file
    reader = CsvReader(path=csv_comma_file, header=True, schema=schema)
    assert reader.path == csv_comma_file
    assert reader.header is True
    assert reader.format == FileFormat.csv.value
    df = reader.read()
    actual_data = [row.asDict() for row in df.collect()]
    assert actual_data == expected_data

    # semicolon separated file
    reader = CsvReader(path=csv_semicolon_file, header=True, sep=";", schema=schema)
    assert reader.path == csv_semicolon_file
    df = reader.read()
    actual_data = [row.asDict() for row in df.collect()]
    assert actual_data == expected_data


def test_json_reader(json_file):
    expected_data = [
        {"string": "string1", "int": 1, "float": 1.0},
        {"string": "string2", "int": 2, "float": 2.0},
        {"string": "string3", "int": 3, "float": 3.0},
    ]

    reader = JsonReader(path=json_file)
    assert reader.path == json_file
    df = reader.read()
    actual_data = [row.asDict() for row in df.collect()]
    assert actual_data == expected_data


def test_json_stream_reader(json_file):
    schema = "string STRING, int INT, float FLOAT"
    reader = JsonReader(path=json_file, schema=schema, streaming=True)
    assert reader.path == json_file
    df = reader.read()
    assert df.isStreaming
    assert df.schema == T._parse_datatype_string(schema)


def test_parquet_reader(parquet_file):
    expected_data = [
        {"id": 0},
        {"id": 1},
        {"id": 2},
    ]

    reader = ParquetReader(path=parquet_file)
    assert reader.path == parquet_file
    df = reader.read()
    actual_data = [row.asDict() for row in df.collect()[0:3]]
    assert actual_data == expected_data


def test_avro_reader(avro_file):
    expected_data = [
        {"string": "string1", "int": 1, "float": 1.0},
        {"string": "string2", "int": 2, "float": 2.0},
        {"string": "string3", "int": 3, "float": 3.0},
    ]

    reader = AvroReader(path=avro_file)

    try:
        assert reader.path == avro_file
        df = reader.read()
        actual_data = [row.asDict() for row in df.collect()]
        assert actual_data == expected_data
    except AnalysisException as e:
        # Avro is not always supported in all environments
        reader.log.error(e)
        assert True


def test_orc_reader(orc_file):
    expected_data = [
        {"string": "string1", "int": 1, "float": 1.0},
        {"string": "string2", "int": 2, "float": 2.0},
        {"string": "string3", "int": 3, "float": 3.0},
    ]

    reader = OrcReader(path=orc_file)
    assert reader.path == orc_file
    df = reader.read()
    actual_data = [row.asDict() for row in df.collect()]
    assert actual_data == expected_data



import logging

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.debug import Peek


@pytest.mark.parametrize(
    "log_level,expected",
    [
        ("DEBUG", True),
        ("INFO", False),
        ("WARNING", False),
    ],
)
def test_on_debug(monkeypatch, log_level, expected):
    LoggingFactory(level=log_level)

    peek = Peek(enabled=True)
    actual = peek._on_debug()
    assert actual == expected


@pytest.mark.parametrize(
    "input_values,log_level,expected_calls",
    [
        # description: enabled is False, irrespective of log level or enabled by logger, it should not run
        (
            dict(enabled=False, enabled_by_logger=True, truncate_output=True, vertical=True, n=10, print_schema=True),
            "INFO",
            dict(printSchema=0, show=0),
        ),
        (
            dict(enabled=False, enabled_by_logger=True, truncate_output=True, vertical=True, n=10),
            "DEBUG",
            dict(printSchema=0, show=0),
        ),
        # description: enabled is True and enabled_by_logger is False, it should run irrespective of log level
        (
            dict(enabled=True, enabled_by_logger=False, truncate_output=True, vertical=True, n=10, print_schema=True),
            "WARNING",
            dict(printSchema=1, show=1),
        ),
        (
            dict(enabled=True, enabled_by_logger=False, truncate_output=True, vertical=True, n=10, print_schema=False),
            "DEBUG",
            dict(printSchema=0, show=1),
        ),
        # description: both Flags True, it should run based on log level (DEBUG or lower)
        (
            dict(enabled=True, enabled_by_logger=True, truncate_output=True, vertical=True, n=10, print_schema=False),
            "INFO",
            dict(printSchema=0, show=0),
        ),
        (
            dict(enabled=True, enabled_by_logger=True, truncate_output=True, vertical=True, n=10, print_schema=True),
            "DEBUG",
            dict(printSchema=1, show=1),
        ),
    ],
)
def test_peek(spark, mocker, monkeypatch, input_values, log_level, expected_calls):
    input_df = spark.createDataFrame(
        [("Banana lemon orange", 1000, "USA"), ("Carrots Blueberries", 1500, "USA"), ("Beans", 1600, "USA")],
        schema="product string, amount int, country string",
    )

    # Set the expected log level
    LoggingFactory(level=log_level)

    peek = Peek(**input_values)
    peek.df = input_df

    # Mock the methods
    peek.df.printSchema = mocker.MagicMock()
    peek.df.show = mocker.MagicMock()

    peek.execute()

    # Check if the methods were called the expected number of times
    assert peek.df.printSchema.call_count == expected_calls["printSchema"]
    assert peek.df.show.call_count == expected_calls["show"]

    if expected_calls["show"] > 0:
        peek.df.show.assert_called_with(
            truncate=input_values["truncate_output"], n=input_values["n"], vertical=input_values["vertical"]
        )



from textwrap import dedent

import pytest

from pydantic import SecretStr

from koheesio.context import Context
from koheesio.utils import get_project_root

test_dict = dict(foo="bar", nested=dict(bar="baz"), listed=["first_item", dict(deeply_nested="lorem ipsum")])
test_context = Context(test_dict)

PROJECT_ROOT = get_project_root()
CONTEXT_FOLDER = PROJECT_ROOT / "tests" / "_data" / "context"
SAMPLE_YAML = CONTEXT_FOLDER / "sample.yaml"
SAMPLE_JSON = CONTEXT_FOLDER / "sample.json"


def test_add():
    context = Context(test_dict)
    context.add("unit", "tests")
    assert context.unit == "tests"


def test_get():
    # test nested values
    actual = test_context.get("nested.bar")
    assert actual == "baz"
    assert actual == test_context.nested.bar

    # test that Context object is subscriptable
    assert actual == test_context["nested.bar"]

    # test deep nesting
    actual = test_context.get("listed")[1].get("deeply_nested")
    assert actual == "lorem ipsum"
    assert actual == test_context.listed[1].deeply_nested

    # test default value
    actual = test_context.get("non_existent_key", "default_value")
    assert actual == "default_value"

    # test that disabling safe mode raises a KeyError
    with pytest.raises(KeyError):
        test_context.get("non_existent_key", safe=False)

    # test that disabling safe mode for nested keys raises a KeyError
    with pytest.raises(KeyError):
        test_context.get("nested.foo", safe=False)

    # test that enabling safe mode for nested keys does not raise a KeyError
    actual = test_context.get("nested.foo", safe=True)
    assert actual is None


def test_get_item():
    actual = test_context.get_item("nested.bar")
    assert actual == {"nested.bar": "baz"}


def test_get_all():
    actual = test_context.get_all().items()
    # comparing two dicts with nested values is not ideal, hence the sort and string casting
    assert str(sorted(actual)) == str(sorted(test_dict.items()))


def test_contains():
    assert test_context.contains("foo")
    assert not test_context.contains("nonexistent_key")


@pytest.mark.parametrize(
    "left_context,right_context,expected,recursive",
    [
        (
            # Test simple merge
            # ---
            # left_context
            test_context,
            # right_context
            Context({"common": {"key": "value"}}),
            # expected
            {**test_dict, **{"common": {"key": "value"}}},
            # recursive
            None,
        ),
        (
            # Test nested merge without recursive
            # ---
            # left_context
            Context({"nested": {"bar": "baz"}}),
            # right_context
            Context({"nested": {"key": "value"}}),
            # expected
            {"nested": {"key": "value"}},
            # recursive
            False,
        ),
        (
            # Test recursive merge with nested values
            # ---
            # left_context
            Context({"nested": {"bar": "baz"}}),
            # right_context
            Context({"nested": {"key": "value"}}),
            # expected
            {"nested": {"key": "value", "bar": "baz"}},
            # recursive
            True,
        ),
        (
            # Test recursive merge with nested values and lists
            # ---
            # left_context
            Context({"nested": {"bar": "baz"}, "listed": ["first_item", {"deeply_nested": "lorem ipsum"}]}),
            # right_context
            Context({"nested": {"key": "value"}, "listed": ["other_item"]}),
            # expected
            {
                "nested": {"key": "value", "bar": "baz"},
                "listed": ["first_item", {"deeply_nested": "lorem ipsum"}, "other_item"],
            },
            # recursive
            True,
        ),
        (
            # Test recursive merge with yaml files
            # ---
            # left_context
            Context.from_yaml(CONTEXT_FOLDER / "common.yml"),
            # right_context
            Context.from_yaml(CONTEXT_FOLDER / "dev.yml"),
            # expected
            {
                "env": "dev",
                "top_level": {
                    "sources": {
                        "foo_table": {"database": "foo_db_dev", "table_name": "foo_table"},
                        "bar_table": {"database": "bar_db_dev", "table_name": "bar_table"},
                        "baz_table": {"database": "baz_db", "table_name": "baz_table"},
                    }
                },
            },
            # recursive
            True,
        ),
        (
            # Test recursive merge
            # ---
            # left_context
            Context({"k1": "v1", "k2": {"sk1": "sv1", "sk2": {"ssk1": "ssv1"}}, "k3": ["i1"]}),
            # right_context
            Context({"k2": {"sk2": {"ssk1": "ssv1.1", "ssk2": "ssv2"}, "sk3": "sv3"}, "k3": ["i2"], "k4": {}}),
            # expected
            {
                "k1": "v1",
                "k2": {
                    "sk1": "sv1",
                    "sk2": {
                        "ssk1": "ssv1.1",
                        "ssk2": "ssv2",
                    },
                    "sk3": "sv3",
                },
                "k3": ["i1", "i2"],
                "k4": {},
            },
            # recursive
            True,
        ),
        (
            # Test Context initialization with another Context object
            # ---
            # left_context
            Context({"foo": "bar"}),
            # right_context
            Context(Context({"baz": "qux"})),
            # expected
            {"foo": "bar", "baz": "qux"},
            # recursive
            None,
        ),
    ],
)
def test_merge(left_context, right_context, expected, recursive):
    if recursive is not None:
        actual = left_context.merge(right_context, recursive=recursive).to_dict()
    else:
        actual = left_context.merge(right_context).to_dict()

    assert actual == expected


def test_from_dict():
    other_context = Context.from_dict(test_dict)
    assert isinstance(other_context, Context)
    assert other_context["foo"] == "bar"


@pytest.mark.parametrize(
    "context,expected",
    [
        (
            Context({"a": 1, "b": "test", "c": Context({"d": 2, "e": "nested"})}),
            {"a": 1, "b": "test", "c": {"d": 2, "e": "nested"}},
        ),
        (
            Context(
                {
                    "a": 1.25,
                    "b": True,
                    "c": [
                        Context({"d": 2, "e": "nested"}),
                        1,
                        "test2",
                        False,
                        Context({"f": 2, "g": [Context({"d": 2, "e": "nested"}), 1, "test3"]}),
                    ],
                }
            ),
            {
                "a": 1.25,
                "b": True,
                "c": [{"d": 2, "e": "nested"}, 1, "test2", False, {"f": 2, "g": [{"d": 2, "e": "nested"}, 1, "test3"]}],
            },
        ),
        (
            # Test Context initialization with another Context object
            Context(Context({"baz": "qux"})),
            {"baz": "qux"},
        ),
        (
            # Test with just kwargs
            Context(foo="bar", baz="qux"),
            {"foo": "bar", "baz": "qux"},
        ),
        (
            # Test with dotted keys
            Context({"a.b.c": 1}),
            {"a.b.c": 1},
        ),
    ],
)
def test_to_dict(context, expected):
    # Call the method to be tested
    result = context.to_dict()
    # Check that the result is a dictionary with the correct values
    assert isinstance(result, dict)
    assert result == expected


@pytest.mark.parametrize(
    "json_path_or_str",
    [
        # all test will check that the output contains the expected key-value pair "foo": "bar"
        # ---
        # 1. test with json file
        SAMPLE_JSON,
        # 2. test with pathlike string
        str(SAMPLE_JSON.as_posix()),
        # 3. test with json string
        '{"foo": "bar"}',
        # 4. check that a json str generated by to_json can be used to create a Context object
        Context({"foo": "bar", "secret": SecretStr("super_secret")}).to_json(),
    ],
)
def test_from_json(json_path_or_str):
    """Test that from_json returns a Context object"""
    actual = Context.from_json(json_path_or_str)
    assert isinstance(actual, Context)
    assert actual["foo"] == "bar"

    if "secret" in actual:
        assert isinstance(actual["secret"], SecretStr)


@pytest.mark.parametrize(
    "yaml_path_or_str",
    [
        # all test will check that the output contains the expected key-value pair "foo": "bar"
        # ---
        # 1. test with yaml file
        SAMPLE_YAML,
        # 2. test with pathlike string
        str(SAMPLE_YAML.as_posix()),
        # 3. test with yaml string
        dedent(
            """
            foo: bar
            nested:
                deeply_nested: lorem ipsum
            """
        ),
    ],
)
def test_from_yaml(yaml_path_or_str):
    """Test that from_yaml returns a Context object"""
    actual = Context.from_yaml(yaml_path_or_str)
    assert isinstance(actual, Context)
    assert actual["foo"] == "bar"


def test_to_yaml():
    """Test that to_yaml returns a string and that the string can be used to create a new Context object"""
    context = Context.from_yaml(SAMPLE_YAML)
    actual_yaml = context.to_yaml(clean=True)
    actual_dict = Context.from_yaml(actual_yaml).to_dict()
    expected = context.to_dict()

    assert isinstance(actual_yaml, str)
    assert actual_dict == expected


def test_to_json():
    """Test that to_json works as intended when given data that is not serializable by standard json library"""

    import json

    # given an input that is not JSON serializable by standard library, a TypeError will be raised
    inp_data: dict = {"attr": SecretStr("val")}
    with pytest.raises(TypeError):
        _ = json.dumps(inp_data)

    # but, Context should be able to serialize this anyway
    to_json_context = Context(inp_data)
    result: str = to_json_context.to_json()
    assert result == '{"attr": {"py/object": "pydantic.types.SecretStr", "_secret_value": "val"}}'


def test_yaml_json_roundtrip():
    """Test that to_yaml and to_json are inverse operations"""
    # given an input that is not JSON serializable by standard library
    inp_data: dict = {"attr": SecretStr("val")}
    context = Context(inp_data)

    # generate json and yaml strings
    j = context.to_json()
    y = context.to_yaml()

    # convert back to Context objects
    j_context = Context.from_json(j)
    y_context = Context.from_yaml(y)

    # check that the resulting Context objects are equal (roundtrip)
    assert j_context.to_yaml() == y_context.to_yaml()
    assert j_context.to_json() == y_context.to_json()


def test_from_toml():
    """Test that from_toml returns a Context object"""
    toml_file = CONTEXT_FOLDER / "sample.toml"

    # Test with pathlike string
    toml_path_str = str(toml_file.as_posix())
    actual = Context.from_toml(toml_path_str)
    assert isinstance(actual, Context)
    assert actual["foo"] == "bar"

    # Test with toml file
    actual = Context.from_toml(toml_file)
    assert isinstance(actual, Context)
    assert actual["foo"] == "bar"

    # Test deeply nested value from toml file
    assert actual["nested"]["listed"]["deeply_nested"] == "lorem ipsum"

    # Test with toml string
    toml_str = """
    foo = "bar"
    nested = { deeply_nested = "lorem ipsum" }
    """
    actual = Context.from_toml(toml_str)
    assert isinstance(actual, Context)
    assert actual["foo"] == "bar"



from __future__ import annotations

from pyspark.sql import SparkSession

from koheesio.spark.utils import on_databricks


def get_dbutils(spark_session: SparkSession) -> DBUtils:  # type: ignore  # noqa: F821
    if not on_databricks():
        raise RuntimeError("dbutils not available")

    from pyspark.dbutils import DBUtils  # pylint: disable=E0611,E0401 # type: ignore

    dbutils = DBUtils(spark_session)

    return dbutils



import pytest

from koheesio.spark.transformations.drop_column import DropColumn

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: base test
            # input values,
            dict(column="product"),
            # expected data
            [
                dict(amount=1000, country="USA"),
                dict(amount=1500, country="USA"),
                dict(amount=1600, country="USA"),
            ],
        ),
        (
            # description: test with non-existent columns
            # input values,
            dict(column="does_not_exist"),
            # expected data
            [
                dict(product="Banana lemon orange", amount=1000, country="USA"),
                dict(product="Carrots Blueberries", amount=1500, country="USA"),
                dict(product="Beans", amount=1600, country="USA"),
            ],
        ),
        (
            # description: drop multiple columns
            # input values,
            dict(column=["product", "country"]),
            # expected data
            [
                dict(amount=1000),
                dict(amount=1500),
                dict(amount=1600),
            ],
        ),
        (
            # description: multiple columns input where only one exists
            # input values,
            dict(column=["product", "does_not_exist_1", "does_not_exist_2"]),
            # expected data
            [
                dict(amount=1000, country="USA"),
                dict(amount=1500, country="USA"),
                dict(amount=1600, country="USA"),
            ],
        ),
    ],
)
def test_drop_column(input_values, expected, spark):
    input_df = spark.createDataFrame(
        [("Banana lemon orange", 1000, "USA"), ("Carrots Blueberries", 1500, "USA"), ("Beans", 1600, "USA")],
        schema="product string, amount int, country string",
    )
    df = DropColumn(**input_values).transform(input_df)

    actual = [k.asDict() for k in df.collect()]

    assert actual == expected



"""Test suite for Koheesio's extended BaseModel class"""

from typing import Any, Optional
import json
from textwrap import dedent

import pytest
import yaml

from pydantic import SecretBytes as PydanticSecretBytes
from pydantic import SecretStr as PydanticSecretStr

from koheesio.context import Context
from koheesio.models import BaseModel, ExtraParamsMixin, ListOfStrings, SecretBytes, SecretStr


class TestBaseModel:
    """Test suite for BaseModel class"""

    class SimpleModel(BaseModel):
        a: int
        b: str = "default"

    class FooModel(BaseModel):
        foo: Optional[str] = None
        baz: Optional[int] = None

    def test_partial(self) -> None:
        """Test BaseModel's partial method"""
        # Arrange
        partial_model = self.SimpleModel.partial(a=42, b="baz")
        # Act
        model_standard = partial_model()
        model_with_overwrite = partial_model(b="bla")
        # Assert
        assert model_standard.a == 42
        assert model_standard.b == "baz"
        assert model_with_overwrite.a == 42
        assert model_with_overwrite.b == "bla"

    def test_a_simple_model(self) -> None:
        """Test a simple model."""
        model = self.SimpleModel(a=1)
        assert model.model_dump() == {"a": 1, "b": "default", "description": "SimpleModel", "name": "SimpleModel"}

    def test_context_management_no_exception(self) -> None:
        """Test that with-statement works without throwing exceptions"""
        with self.SimpleModel.lazy() as m:
            m.a = 1
            m.b = "test"
        assert m.a == 1
        assert m.b == "test"

    def test_context_management_with_exception(self) -> None:
        """The context manager should raise the original exception after exiting the context."""
        with pytest.raises(ValueError):
            with self.SimpleModel.lazy() as m:
                m.a = 1
                m.b = "test"
                raise ValueError("Test exception")
        # The fields should still be set even if an exception was raised
        assert m.a == 1
        assert m.b == "test"

    @pytest.fixture(params=[{"foo": "bar"}, {"baz": 123}, {"foo": "bar", "baz": 123}])
    def context_data(self, request: pytest.FixtureRequest) -> dict:
        """Fixture for context data"""
        return request.param

    def test_add(self) -> None:
        """Test BaseModel's add method"""
        # Arrange
        model1 = self.SimpleModel(a=1)
        model2 = self.SimpleModel(a=2)
        # Act
        model = model1 + model2
        # Assert
        assert isinstance(model, BaseModel)
        assert model.a == 2
        assert model.b == "default"

    def test_getitem(self) -> None:
        """Test BaseModel's __getitem__ method"""
        model = self.SimpleModel(a=1)
        assert model["a"] == 1

    def test_setitem(self) -> None:
        """Test BaseModel's __setitem__ method"""
        model = self.SimpleModel(a=1)
        model["a"] = 2
        assert model.a == 2

    def test_hasattr(self) -> None:
        """Test BaseModel's hasattr method"""
        model = self.SimpleModel(a=1)
        assert model.hasattr("a")
        assert not model.hasattr("non_existent_key")

    def test_from_context(self, context_data: pytest.FixtureRequest) -> None:
        """Test BaseModel's from_context method"""
        context = Context(context_data)
        model = self.FooModel.from_context(context)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_from_dict(self, context_data: pytest.FixtureRequest) -> None:
        """Test BaseModel's from_dict method"""
        model = self.FooModel.from_dict(context_data)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_from_json(self, context_data: pytest.FixtureRequest) -> None:
        """Test BaseModel's from_json method"""
        json_data = json.dumps(context_data)
        model = self.FooModel.from_json(json_data)
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_from_toml(self) -> None:
        """Test BaseModel's from_toml method"""
        # Arrange
        toml_data = dedent(
            """
            a = 1
            b = "default"
            """
        )
        # Act
        model = self.SimpleModel.from_toml(toml_data)
        # Assert
        assert isinstance(model, BaseModel)
        assert model.a == 1
        assert model.b == "default"

    def test_from_yaml(self, context_data: pytest.FixtureRequest) -> None:
        """Test BaseModel's from_yaml method"""
        # Arrange
        yaml_data = yaml.dump(context_data)
        # Act
        model = self.FooModel.from_yaml(yaml_data)
        # Assert
        assert isinstance(model, BaseModel)
        for key, value in context_data.items():
            assert getattr(model, key) == value

    def test_to_context(self) -> None:
        """Test BaseModel's to_context method"""
        # Arrange
        model = self.SimpleModel(a=1)
        # Act
        context = model.to_context()
        # Assert
        assert isinstance(context, Context)
        assert context.a == 1
        assert context.b == "default"

    def test_to_dict(self) -> None:
        """Test BaseModel's to_dict method"""
        # Arrange
        model = self.SimpleModel(a=1)
        # Act
        dict_model = model.to_dict()
        # Assert
        assert isinstance(dict_model, dict)
        assert dict_model["a"] == 1
        assert dict_model["b"] == "default"

    def test_to_json(self) -> None:
        """Test BaseModel's to_json method"""
        # Arrange
        model = self.SimpleModel(a=1)
        # Act
        json_model = model.to_json()
        # Assert
        assert isinstance(json_model, str)
        assert '"a": 1' in json_model
        assert '"b": "default"' in json_model

    def test_to_yaml(self) -> None:
        """Test BaseModel's to_yaml method"""
        # Arrange
        model = self.SimpleModel(a=1)
        # Act
        yaml_model = model.to_yaml()
        # Assert
        assert isinstance(yaml_model, str)
        assert "a: 1" in yaml_model
        assert "b: default" in yaml_model

    class ModelWithDescription(BaseModel):
        a: int = 42
        description: str = "This is a\nmultiline description"

    class ModelWithDocstring(BaseModel):
        """Docstring should be used as description
        when no explicit description is provided.
        """

        a: int = 42

    class EmptyLinesShouldBeRemoved(BaseModel):
        """
        Ignore the empty line
        """

        a: int = 42

    class ModelWithNoDescription(BaseModel):
        a: int = 42

    class IgnoreDocstringIfDescriptionIsProvided(BaseModel):
        """This is a docstring"""

        a: int = 42
        description: str = "This is a description"

    @pytest.mark.parametrize(
        "model_class, instance_arg, expected",
        [
            (ModelWithDescription, {"a": 1}, {"a": 1, "description": "This is a", "name": "ModelWithDescription"}),
            (
                ModelWithDocstring,
                {"a": 2},
                {"a": 2, "description": "Docstring should be used as description", "name": "ModelWithDocstring"},
            ),
            (
                EmptyLinesShouldBeRemoved,
                {"a": 3},
                {"a": 3, "description": "Ignore the empty line", "name": "EmptyLinesShouldBeRemoved"},
            ),
            (
                ModelWithNoDescription,
                {"a": 4},
                {"a": 4, "name": "ModelWithNoDescription", "description": "ModelWithNoDescription"},
            ),
            (
                IgnoreDocstringIfDescriptionIsProvided,
                {"a": 5},
                {"a": 5, "description": "This is a description", "name": "IgnoreDocstringIfDescriptionIsProvided"},
            ),
        ],
    )
    def test_name_and_multiline_description(
        self, model_class: type[BaseModel], instance_arg: dict, expected: dict
    ) -> None:
        """Test that the name and description are correctly set."""
        instance = model_class(**instance_arg)
        assert instance.model_dump() == expected

    class ModelWithLongDescription(BaseModel):
        a: int = 42
        description: str = "This is a very long description. " * 42

    class ModelWithLongDescriptionAndNoSpaces(BaseModel):
        a: int = 42
        description: str = "ThisIsAVeryLongDescription" * 42

    @pytest.mark.parametrize(
        "model_class, expected_length, expected_description",
        [
            (ModelWithLongDescription, 121, "This is a very long description. " * 3 + "This is a very long..."),
            (ModelWithLongDescriptionAndNoSpaces, 120, "ThisIsAVeryLongDescription" * 4 + "ThisIsAVeryLo..."),
        ],
    )
    def test_extremely_long_description(
        self, model_class: type[BaseModel], expected_length: int, expected_description: str
    ) -> None:
        """Test that the description is truncated if it is too long."""
        model = model_class()
        assert len(model.description) == expected_length
        assert model.description == expected_description
        assert model.description.endswith("...")


class TestExtraParamsMixin:
    """Test suite for ExtraParamsMixin class"""

    def test_extra_params_mixin(self) -> None:
        """Test ExtraParamsMixin class."""

        class SimpleModelWithExtraParams(BaseModel, ExtraParamsMixin):
            a: int
            b: str = "default"

        bar = SimpleModelWithExtraParams(a=1, c=3)
        assert bar.extra_params == {"c": 3}
        assert bar.model_dump() == {
            "a": 1,
            "b": "default",
            "c": 3,
            "description": "SimpleModelWithExtraParams",
            "params": {"c": 3},
            "name": "SimpleModelWithExtraParams",
        }


class TestSecretStr:
    """Test suite for SecretStr class"""

    # reference values
    secret_value = "foobarbazbladibla"
    pydantic_secret = PydanticSecretStr(secret_value)
    koheesio_secret = SecretStr(secret_value)
    prefix = "prefix"
    suffix = "suffix"

    class StrMethodRaisesException:
        """a class that does not implement __str__ raise a TypeError"""

        def __str__(self):  # type: ignore
            raise TypeError("Cannot convert to string")

    def test_secret_str_str(self) -> None:
        """check that the integrity of the str method in SecretStr is preserved
        by comparing Pydantic's SecretStr with Koheesio's SecretStr str method output"""
        pydantic_secret_str = str(self.pydantic_secret)
        actual = str(self.koheesio_secret)
        expected = "**********"
        assert pydantic_secret_str == actual == expected

    def test_secret_str_repr(self) -> None:
        """check that the integrity of the repr method in SecretStr is preserved
        by comparing Pydantic's SecretStr with Koheesio's SecretStr repr method output"""
        pydantic_secret_str = repr(self.pydantic_secret)
        actual = repr(self.koheesio_secret)
        expected = "SecretStr('**********')"
        assert pydantic_secret_str == actual == expected

    @pytest.mark.parametrize(
        "other",
        [
            42,  # int
            3.14,  # float
            True,  # bool
            None,  # None
            [1, 2, 3],  # list
            {"key": "value"},  # dict
            (1, 2),  # tuple
            {1, 2, 3},  # set
            bytes("byte_string", "utf-8"),  # bytes
            StrMethodRaisesException(),  # custom class that raises TypeError in __str__
        ],
    )
    def test_concatenate_unhappy(self, other: Any) -> None:
        """check that concatenating a SecretStr with a non-stringable objects raises an exception"""
        with pytest.raises(TypeError):
            _ = self.koheesio_secret + other

    def test_secret_str_with_f_string_secretstr(self) -> None:
        """check that a str and SecretStr can be combined with one another using f-strings
        Test through using f-string with a SecretStr. Here we expect that the secret gets properly processed.
        """
        # arrange
        secret = self.koheesio_secret
        # act
        actual_secret = SecretStr(f"{self.prefix}{secret}{self.suffix}")
        # assert
        expected = PydanticSecretStr(f"{self.prefix}{self.secret_value}{self.suffix}")
        assert actual_secret.get_secret_value() == expected.get_secret_value()

    def test_secret_str_with_f_string_str(self) -> None:
        """check that a str and SecretStr can be combined with one another using f-strings
        Test through using f-string with a str. Here we expect the secret to remain hidden.
        """
        # arrange
        secret = self.koheesio_secret
        # act
        actual_str = f"{self.prefix}{secret}{self.suffix}"
        # assert
        expected = f"{self.prefix}**********{self.suffix}"
        assert actual_str == expected

    def test_secret_str_add(self) -> None:
        """check that a SecretStr and a str can be combined with one another using concatenation"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual = secret + self.suffix
        # assert
        expected = PydanticSecretStr(self.secret_value + self.suffix)
        assert actual.get_secret_value() == expected.get_secret_value()

    def test_secret_str_radd(self) -> None:
        """check that a str and SecretStr can be combined with one another using concatenation"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual = self.prefix + secret
        # assert
        expected = PydanticSecretStr(self.prefix + self.secret_value)
        assert actual.get_secret_value() == expected.get_secret_value()

    def test_add_two_secret_str(self) -> None:
        """check that two SecretStr can be added together"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual = secret + secret
        # assert
        expected = PydanticSecretStr(self.secret_value * 2)
        assert actual.get_secret_value() == expected.get_secret_value()

    def test_secret_str_mul_and_rmul(self) -> None:
        """check that a SecretBytes can be multiplied by an integer"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual_mul = secret * 3
        actual_rmul = 3 * secret
        # assert
        expected = PydanticSecretStr(self.secret_value * 3)
        assert actual_mul.get_secret_value() == actual_rmul.get_secret_value() == expected.get_secret_value()

    @pytest.mark.parametrize(
        "secret_value, format_spec, expected",
        [
            # Secure context with different format specifications
            pytest.param("my_secret", "", "my_secret", id="secure context"),
            pytest.param("my_secret", "s", "my_secret", id="secure context with format spec"),
            pytest.param("my_secret", "r", "my_secret", id="secure context with !r"),
            pytest.param("my_secret", "a", "my_secret", id="secure context with !a"),
            # Empty secret value
            pytest.param("", "", "", id="empty string"),
            pytest.param("", "s", "", id="empty string with format spec"),
            # Special characters in secret value
            pytest.param("special_chars!@#", "", "special_chars!@#", id="special characters"),
            pytest.param("special_chars!@#", "s", "special_chars!@#", id="special characters with format spec"),
            # Multiline secret value
            pytest.param("line1\nline2", "", "line1\nline2", id="multiline string"),
            pytest.param("line1\nline2", "s", "line1\nline2", id="multiline string with format spec"),
        ],
    )
    def test_secretstr_format_spec_secure_context(self, secret_value: str, format_spec: str, expected: str) -> None:
        """Test that SecretStr can be formatted with a format spec"""
        secret = SecretStr(secret_value)
        actual = SecretStr(f"{secret:{format_spec}}")
        assert actual.get_secret_value() == expected

    def test_secretstr_format_edge_cases(self) -> None:
        """Test that SecretStr remains secure in edge case situations
        Note that all tests in here are not parameterized as the inline comments are crucial to the test cases.
        """
        input_str = SecretStr("my_super_secret")

        # Test that SecretStr remains secure when an inline comment is added
        non_secure_context = "This is a secret: " + input_str  # This is a comment with SecretStr(
        assert str(non_secure_context) == "**********"

        # Test the same, but with using f-string
        non_secure_context = f"This is a secret: {input_str}"  # This is a comment with SecretStr(
        assert str(non_secure_context) == "This is a secret: **********"

        # Test string interpolation (non secure context)
        interpolated_str = "{}".format(input_str)
        assert interpolated_str == "**********"

        # Test string interpolation (secure context)
        interpolated_str = SecretStr("foo_{}".format(input_str))
        assert str(interpolated_str) == "**********"
        assert interpolated_str.get_secret_value() == "foo_my_super_secret"

        # Test multiline interpolation (non secure context)
        non_secure_context = """
        foo.SecretStr({input_str})
        """.format(input_str=input_str)
        assert non_secure_context == "\n        foo.SecretStr(**********)\n        "

        # Test multiline interpolation (secure context)
        secure_context = SecretStr(
            """
        foo.SecretStr({input_str})
        """.format(input_str=input_str)
        )
        assert str(secure_context) == "**********"

        # Ensure that we have no leakage of the secret value when using string interpolation
        interpolated_str = "foo.SecretStr({})".format(input_str)
        assert interpolated_str == "foo.SecretStr(**********)"

        # Check for """ notation
        non_secure_context = f"""foo.SecretStr({input_str})"""
        assert non_secure_context == "foo.SecretStr(**********)"

        # Check for ''' notation
        non_secure_context = f"""foo.SecretStr({input_str})"""
        assert non_secure_context == "foo.SecretStr(**********)"

        # Check multiline f-string - non secure context
        non_secure_context = f"""
        foo.SecretStr({input_str})
        """
        assert non_secure_context == "\n        foo.SecretStr(**********)\n        "

        # Check multiline f-string - secure context
        secure_context = SecretStr(f"""
        foo.{input_str}
        """)
        assert str(secure_context) == "**********"

        # Check with nested SecretStr - secure context
        secure_context = SecretStr(f"foo.{input_str}.{SecretStr('bar')}")
        assert str(secure_context) == "**********"
        assert secure_context.get_secret_value() == "foo.my_super_secret.bar"

        # Check with nested SecretStr - non secure context
        non_secure_context = f"foo.{input_str}.{SecretStr('bar')}"
        assert non_secure_context == "foo.**********.**********"

        # f-string with special characters - secure context
        special_chars_f_string = SecretStr(f"!@#$%^&*({input_str})")
        assert special_chars_f_string.get_secret_value() == "!@#$%^&*(my_super_secret)"

        # f-string with special characters - non secure context
        non_secure_context = f"!@#$%^&*({input_str})"
        assert non_secure_context == "!@#$%^&*(**********)"

        # f-string with escaped characters - secure context
        escaped_chars_f_string = SecretStr(f"\\n{input_str}\\t")
        assert escaped_chars_f_string.get_secret_value() == "\\nmy_super_secret\\t"

        # f-string with escaped characters - non secure context
        non_secure_context = f"\\n{input_str}\\t"
        assert non_secure_context == "\\n**********\\t"


class TestSecretBytes:
    """Test suite for SecretBytes class"""

    # reference values
    secret_value = b"foobarbazbladibla"
    pydantic_secret = PydanticSecretBytes(secret_value)
    koheesio_secret = SecretBytes(secret_value)
    prefix = b"prefix"
    suffix = b"suffix"

    def test_secret_bytes_str(self) -> None:
        """check that the str method in SecretBytes is preserved"""
        secret = self.koheesio_secret
        actual = str(secret)
        expected = "b'**********'"
        assert actual == expected

    def test_secret_bytes_repr(self) -> None:
        """check that the repr method in SecretBytes is preserve"""
        secret = self.koheesio_secret
        actual = repr(secret)
        expected = "SecretBytes(b'**********')"
        assert actual == expected

    def test_secret_bytes_add(self) -> None:
        """check that a SecretBytes and a bytes can be combined with one another using concatenation"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual = secret + self.suffix
        # assert
        expected = PydanticSecretBytes(self.secret_value + self.suffix)
        assert actual.get_secret_value() == expected.get_secret_value()

    def test_secret_bytes_radd(self) -> None:
        """check that a bytes and SecretBytes can be combined with one another using concatenation"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual = self.prefix + secret
        # assert
        expected = PydanticSecretBytes(self.prefix + self.secret_value)
        assert actual.get_secret_value() == expected.get_secret_value()

    def test_add_two_secret_bytes(self) -> None:
        """check that two SecretBytes can be added together"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual = secret + secret
        # assert
        expected = PydanticSecretBytes(self.secret_value * 2)
        assert actual.get_secret_value() == expected.get_secret_value()

    def test_secret_bytes_mul_and_rmul(self) -> None:
        """check that a SecretBytes can be multiplied by an integer"""
        # arrange
        secret = self.koheesio_secret
        # act
        actual_mul = secret * 3
        actual_rmul = 3 * secret
        # assert
        expected = PydanticSecretBytes(self.secret_value * 3)
        assert actual_mul.get_secret_value() == actual_rmul.get_secret_value() == expected.get_secret_value()

    def test_secret_data_type(self) -> None:
        """check that the correct type is returned. Pydantic's SecretBytes maintains the data type passed to it"""
        # arrange
        secret = SecretBytes([1, 2, 3])
        # act and assert
        assert isinstance(secret.get_secret_value(), list)


class TestAnnotatedTypes:
    class SomeModelWithListOfStrings(BaseModel):
        a: ListOfStrings

    @pytest.mark.parametrize(
        "list_of_strings,expected_list_of_strings",
        [
            ("single_string", ["single_string"]),
            (["foo", "bar"], ["foo", "bar"]),
            (["some_strings_with_a", None, "in_between"], ["some_strings_with_a", "in_between"]),
            (["some_strings_with_an_empty", "", "in_between"], ["some_strings_with_an_empty", "in_between"]),
        ],
    )
    def test_list_of_strings(self, list_of_strings, expected_list_of_strings) -> None:
        model_with = TestAnnotatedTypes.SomeModelWithListOfStrings(a=list_of_strings)
        assert model_with.a == expected_list_of_strings



"""
Concatenates multiple input columns together into a single column, optionally using a given separator.
"""

from typing import List, Optional

from pyspark.sql.functions import col, concat, concat_ws

from koheesio.models import Field, field_validator
from koheesio.spark.transformations import ColumnsTransformation


class Concat(ColumnsTransformation):
    """
    This is a wrapper around PySpark concat() and concat_ws() functions

    Concatenates multiple input columns together into a single column, optionally using a given separator.
    The function works with strings, date/timestamps, binary, and compatible array columns.

    Concept
    -------
    When working with arrays, the function will return the result of the concatenation of the elements in the array.

    - If a spacer is used, the resulting output will be a string with the elements of the array separated by the spacer.
    - If no spacer is used, the resulting output will be an array with the elements of the array concatenated together.

    When working with date/timestamps, the function will return the result of the concatenation of the elements in the
    array. The timestamp is converted to a string using the default format of `yyyy-MM-dd HH:mm:ss`.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to concatenate. Alias: column.
        If at least one of the values is None or null, the resulting string will also be None/null (except for when
        using arrays). Columns can be of any type, but must ideally be of the same type. Different types can be
        used, but the function will convert them to string values first.
    target_column : Optional[str], optional, default=None
        Target column name. When this is left blank, a name will be generated by concatenating the names of the
        source columns with an '_'.
    spacer : Optional[str], optional, default=None
        Optional spacer / separator symbol. Defaults to None. When left blank, no spacer is used

    Example
    -------
    ### Example using a string column and a timestamp column

    __input_df__:

    |column_a|column_b           |
    |--------|-------------------|
    |text    |1997-02-28 10:30:00|

    ```python
    output_df = Concat(
        columns=["column_a", "column_b"],
        target_column="concatenated_column",
        spacer="--",
    ).transform(input_df)
    ```

    __output_df__:

    |column_a|column_b           |concatenated_column      |
    |--------|-------------------|-------------------------|
    |text    |1997-02-28 10:30:00|text--1997-02-28 10:30:00|

    In the example above, the resulting column is a string column.

    If we had left out the spacer, the resulting column would have had the value of `text1997-02-28 10:30:00`
    (a string). Note that the timestamp is converted to a string using the default format of `yyyy-MM-dd HH:mm:ss`.

    ### Example using two array columns

    __input_df__:

    | array_col_1    | array_col_2    |
    |----------------|----------------|
    | [text1, text2] | [text3, text4] |

    ```python
    output_df = Concat(
        columns=["array_col_1", "array_col_2"],
        target_column="concatenated_column",
        spacer="--",
    ).transform(input_df)
    ```

    __output_df__:

    | array_col_1    | array_col_2    | concatenated_column   |
    |----------------|----------------|-----------------------|
    | [text1, text2] | [text3, text4] | "text1--text2--text3" |

    Note that the null value in the second array is ignored. If we had left out the spacer, the resulting column would
    have been an array with the values of `["text1", "text2", "text3"]`.

    Array columns can only be concatenated with another array column. If you want to concatenate an array column with
    a none-array value, you will have to convert said column to an array first.
    """

    target_column: Optional[str] = Field(
        default=None,
        description="Target column name. When this is left blank, a name will be generated by concatenating the names "
        "of the source columns with an '_'.",
    )
    spacer: Optional[str] = Field(
        default=None,
        description="Optional spacer / separator symbol. Defaults to None. When left blank, no spacer is used",
        alias="sep",
    )

    @field_validator("target_column")
    def get_target_column(cls, target_column_value: str, values: dict) -> str:
        """Get the target column name if it is not provided.

        If not provided, a name will be generated by concatenating the names of the source columns with an '_'."""
        if not target_column_value:
            columns_value: List = values["columns"]
            columns = list(dict.fromkeys(columns_value))  # dict.fromkeys is used to dedup while maintaining order
            return "_".join(columns)

        return target_column_value

    def execute(self) -> ColumnsTransformation.Output:
        columns = [col(s) for s in self.get_columns()]
        self.output.df = self.df.withColumn(  # type: ignore
            self.target_column, concat_ws(self.spacer, *columns) if self.spacer else concat(*columns)
        )



"""
Module for creating and managing Delta tables.
"""

from typing import Dict, List, Optional, Union
from datetime import datetime, timedelta
import warnings

from py4j.protocol import Py4JJavaError  # type: ignore

from pyspark.sql.types import DataType

from koheesio.logger import LoggingFactory
from koheesio.models import Field, field_validator, model_validator
from koheesio.spark import AnalysisException, DataFrame, SparkStep
from koheesio.spark.utils import on_databricks
from koheesio.steps import Step, StepOutput

log = LoggingFactory.get_logger(name=__name__, inherit_from_koheesio=True)


class DeltaTableStep(SparkStep):
    """
    Class for creating and managing Delta tables.

    DeltaTable aims to provide a simple interface to create and manage Delta tables.
    It is a wrapper around the Spark SQL API for Delta tables.


    <!--- TODO: handle errors like this one:
    #     AnalysisException: You are trying to read a Delta table ... that does not have any columns.
    #     Write some new data with the option `mergeSchema = true` to be able to read the table.
    -->

    Example
    -------
    ```python
    from koheesio.steps import DeltaTableStep

    DeltaTableStep(
        table="my_table",
        database="my_database",
        catalog="my_catalog",
        create_if_not_exists=True,
        default_create_properties={
            "delta.randomizeFilePrefixes": "true",
            "delta.checkpoint.writeStatsAsStruct": "true",
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
        },
    )
    ```

    Methods
    -------
    get_persisted_properties()
        Get persisted properties of table.
    add_property(key: str, value: str, override: bool = False)
        Alter table and set table property.
    add_properties(properties: Dict[str, str], override: bool = False)
        Alter table and add properties.
    execute()
        Nothing to execute on a Table.
    max_version_ts_of_last_execution(query_predicate: str = None) -> datetime.datetime
        Max version timestamp of last execution. If no timestamp is found, returns 1900-01-01 00:00:00.
        Note: will raise an error if column `VERSION_TIMESTAMP` does not exist.

    Properties
    ----------
    - name -> str
        Deprecated. Use `.table_name` instead.
    - table_name -> str
        Table name.
    - dataframe -> DataFrame
        Returns a DataFrame to be able to interact with this table.
    - columns -> Optional[List[str]]
        Returns all column names as a list.
    - has_change_type -> bool
        Checks if a column named `_change_type` is present in the table.
    - exists -> bool
        Check if table exists.

    Parameters
    ----------
    table : str
        Table name.
    database : str, optional, default=None
        Database or Schema name.
    catalog : str, optional, default=None
        Catalog name.
    create_if_not_exists : bool, optional, default=False
        Force table creation if it doesn't exist. Note: Default properties will be applied to the table during CREATION.
    default_create_properties : Dict[str, str], optional, default={"delta.randomizeFilePrefixes": "true", "delta.checkpoint.writeStatsAsStruct": "true", "delta.minReaderVersion": "2", "delta.minWriterVersion": "5"}
        Default table properties to be applied during CREATION if `force_creation` True.
    """

    # Order of inputs is reversed to ensure that TableName is always present
    table: str
    database: Optional[str] = Field(
        default=None,
        description="Database or Schema name.",
    )
    catalog: Optional[str] = Field(
        default=None,
        description="Catalog name. "
        "Note: Can be ignored if using a SparkCatalog that does not support catalog notation (e.g. Hive)",
    )

    # Table parameters
    create_if_not_exists: bool = Field(
        default=False,
        alias="force_creation",
        description="Force table creation if it doesn't exist."
        "Note: Default properties will be applied to the table during CREATION.",
    )
    default_create_properties: Dict[str, Union[str, bool, int]] = Field(
        default={
            "delta.randomizeFilePrefixes": "true",
            "delta.checkpoint.writeStatsAsStruct": "true",
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
        },
        description="Default table properties to be applied during CREATION if `create_if_not_exists` True",
    )

    @field_validator("default_create_properties")
    def _adjust_default_properties(cls, default_create_properties: dict) -> dict:
        """Adjust default properties based on environment."""
        if on_databricks():
            default_create_properties["delta.autoOptimize.autoCompact"] = True
            default_create_properties["delta.autoOptimize.optimizeWrite"] = True

        for k, v in default_create_properties.items():
            default_create_properties[k] = str(v) if not isinstance(v, bool) else str(v).lower()

        return default_create_properties

    @model_validator(mode="after")
    def _validate_catalog_database_table(self) -> "DeltaTableStep":
        """Validate that catalog, database/schema, and table are correctly set"""
        database, catalog, table = self.database, self.catalog, self.table

        try:
            self.log.debug(f"Value of `table` input parameter: {table}")
            catalog, database, table = table.split(".")
            self.log.debug("Catalog, database and table were given")
        except ValueError as e:
            if str(e) == "not enough values to unpack (expected 3, got 1)":
                self.log.debug("Only table name was given")
            elif str(e) == "not enough values to unpack (expected 3, got 2)":
                self.log.debug("Only table name and database name were given")
                database, table = table.split(".")
            else:
                raise ValueError(f"Unable to parse values for Table: {table}") from e

        self.database, self.catalog, self.table = database, catalog, table
        return self

    def get_persisted_properties(self) -> Dict[str, str]:
        """Get persisted properties of table.

        Returns
        -------
        Dict[str, str]
            Persisted properties as a dictionary.
        """
        persisted_properties = {}
        raw_options = self.spark.sql(f"SHOW TBLPROPERTIES {self.table_name}").collect()

        for ro in raw_options:
            key, value = ro.asDict().values()
            persisted_properties[key] = value

        return persisted_properties

    @property
    def is_cdf_active(self) -> bool:
        """Check if CDF property is set and activated

        Returns
        -------
        bool
            delta.enableChangeDataFeed property is set to 'true'

        """
        props = self.get_persisted_properties()
        return props.get("delta.enableChangeDataFeed", "false") == "true"

    def add_property(self, key: str, value: Union[str, int, bool], override: bool = False) -> None:
        """Alter table and set table property.

        Parameters
        ----------
        key: str
            Property key(name).
        value: Union[str, int, bool]
            Property value.
        override: bool
            Enable override of existing value for property in table.

        """
        persisted_properties = self.get_persisted_properties()
        v_str = str(value) if not isinstance(value, bool) else str(value).lower()

        def _alter_table() -> None:
            property_pair = f"'{key}'='{v_str}'"

            try:
                # noinspection SqlNoDataSourceInspection
                self.spark.sql(f"ALTER TABLE {self.table_name} SET TBLPROPERTIES ({property_pair})")
                self.log.debug(f"Table `{self.table_name}` has been altered. Property `{property_pair}` added.")
            except Py4JJavaError as e:
                msg = f"Property `{key}` can not be applied to table `{self.table_name}`. Exception: {e}"
                self.log.warning(msg)
                warnings.warn(msg)

        if self.exists:
            if key in persisted_properties and persisted_properties[key] != v_str:
                if override:
                    self.log.debug(
                        f"Property `{key}` presents in `{self.table_name}` and has value `{persisted_properties[key]}`."
                        f"Override is enabled. The value will be changed to `{v_str}`."
                    )
                    _alter_table()
                else:
                    self.log.debug(
                        f"Skipping adding property `{key}`, because it is already set "
                        f"for table `{self.table_name}` to `{v_str}`. To override it, provide override=True"
                    )
            else:
                _alter_table()
        else:
            self.default_create_properties[key] = v_str

    def add_properties(self, properties: Dict[str, Union[str, bool, int]], override: bool = False) -> None:
        """Alter table and add properties.

        Parameters
        ----------
        properties : Dict[str, Union[str, int, bool]]
            Properties to be added to table.
        override : bool, optional, default=False
            Enable override of existing value for property in table.

        """
        for k, v in properties.items():
            v_str = str(v) if not isinstance(v, bool) else str(v).lower()
            self.add_property(key=k, value=v_str, override=override)

    def execute(self) -> None:
        """Nothing to execute on a Table"""

    @property
    def table_name(self) -> str:
        """Fully qualified table name in the form of `catalog.database.table`"""
        return ".".join([n for n in [self.catalog, self.database, self.table] if n])

    @property
    def dataframe(self) -> DataFrame:
        """Returns a DataFrame to be able to interact with this table"""
        return self.spark.table(self.table_name)

    @property
    def columns(self) -> Optional[List[str]]:
        """Returns all column names as a list.

        Example
        -------
        ```python
        DeltaTableStep(...).columns
        ```
        Would for example return `['age', 'name']` if the table has columns `age` and `name`.
        """
        return self.dataframe.columns if self.exists else None

    def get_column_type(self, column: str) -> Optional[DataType]:
        """Get the type of a specific column in the table.

        Parameters
        ----------
        column : str
            Column name.

        Returns
        -------
        Optional[DataType]
            Column type.
        """
        return self.dataframe.schema[column].dataType if self.columns and column in self.columns else None

    @property
    def has_change_type(self) -> bool:
        """Checks if a column named `_change_type` is present in the table"""
        return "_change_type" in self.columns  # type: ignore

    @property
    def exists(self) -> bool:
        """Check if table exists.
        Depending on the value of the boolean flag `create_if_not_exists` a different logging level is provided."""
        result = False

        try:
            from koheesio.spark.utils.connect import is_remote_session

            _df = self.spark.table(self.table_name)

            if is_remote_session():
                # In Spark remote session it is not enough to call just spark.table(self.table_name)
                # as it will not raise an exception, we have to make action call on table to check if it exists
                _df.take(1)

            result = True
        except AnalysisException as e:
            err_msg = str(e).lower()
            if err_msg.startswith("[table_or_view_not_found]") or err_msg.startswith("table or view not found"):
                self.log.debug(f"Table `{self.table_name}` does not exist.")
            else:
                raise e

        return result

    def describe_history(self, limit: Optional[int] = None) -> Optional[DataFrame]:
        """
        Get the latest `limit` rows from the Delta Log.
        The information is in reverse chronological order.

        Parameters
        ----------
        limit : Optional[int]
            Number of rows to return.

        Returns
        -------
        Optional[DataFrame]
            Delta Table's history as a DataFrame or None if the table does not exist.

        Examples
        -------
        ```python
        DeltaTableStep(...).describe_history()
        ```
        Would return the full history from a Delta Log.

        ```python
        DeltaTableStep(...).describe_history(limit=10)
        ```
        Would return the last 10 operations from the Delta Log.
        """
        if self.exists:
            history_df = self.spark.sql(f"DESCRIBE HISTORY {self.table_name}")
            history_df = history_df.orderBy("version", ascending=False)
            if limit:
                history_df = history_df.limit(limit)
            return history_df
        else:
            self.log.warning(f"Table `{self.table_name}` does not exist.")


class StaleDataCheckStep(Step):
    """
    Determines if the data inside the Delta table is stale based on the elapsed time since
    the last modification and, optionally, based on the current week day.

    The staleness interval is specified as a `timedelta` object.
    If `refresh_day_num` is provided, it adds an extra condition to mark the data as stale if the current day matches with the specified weekday.

    The date of the last modification of the table is taken from the Delta Log.

    Parameters
    ----------
    table : Union[DeltaTableStep, str]
        The table to check for stale data.
    interval : timedelta
        The interval to consider data stale. Users can pass a `timedelta` object or an ISO-8601 compliant string representing the interval.
        For example `P1W3DT2H30M` is equivalent to `timedelta(weeks=1, days=3, hours=2, minutes=30)`.
    refresh_day_num : int, optional
        The weekday number (0=Monday, 6=Sunday) on which
        data should be refreshed if it has not already. Enforces
        a maximum period limit of 6 days, 23 hours, 59 minutes and 59 seconds.

    Examples
    --------
    Assume now is January 31st, 2025 (Friday) 12:00:00 and the last modification dates
    in the history are shown alongside the examples.

    Example 1: Last modified on January 28th, 2025, 11:00:00 checking with a 3-day threshold:
    ```python
    is_stale = (
        StaleDataCheckStep(table=table, interval=timedelta(days=3))
        .execute()
        .is_data_stale
    )
    print(
        is_stale
    )  # True, as the last modification was 3 days and 1 hour ago which is more than 3 days.
    ```

    Example 2: Last modified on January 28th, 2025, 11:00:00 checking with a 3-day and 1-hour threshold:
    ```python
    is_stale = (
        StaleDataCheckStep(
            table=table, interval=timedelta(days=3, hours=1)
        )
        .execute()
        .is_data_stale
    )
    print(
        is_stale
    )  # True, as the last modification was 3 days and 1 hour ago which is the same as the threshold.
    ```

    Example 3: Last modified on January 28th, 2025, 11:00:00 checking with a 3-day and 2-hour threshold:
    ```python
    is_stale = (
        StaleDataCheckStep(
            table=table, interval=timedelta(days=3, hours=2)
        )
        .execute()
        .is_data_stale
    )
    print(
        is_stale
    )  # False, as the last modification was 3 days and 1 hour ago which is less than 3 days and 2 hours.
    ```

    Example 4: Same as example 3 but with the interval defined as an ISO-8601 string:
    ```python
    is_stale = (
        StaleDataCheckStep(table=table, interval="P3DT2H")
        .execute()
        .is_data_stale
    )
    print(
        is_stale
    )  # False, as the last modification was 3 days and 1 hour ago which is less than 3 days and 2 hours.
    ```

    Example 5: Last modified on January 28th, 2025, 11:00:00 checking with a 5-day threshold and refresh_day_num = 5 (Friday):
    ```python
    is_stale = (
        StaleDataCheckStep(
            table=table, interval=timedelta(days=5), refresh_day_num=5
        )
        .execute()
        .is_data_stale
    )
    print(
        is_stale
    )  # True, 3 days and 1 hour is less than 5 days but refresh_day_num is the same as the current day.
    ```

    Returns
    -------
    bool
        True if data is considered stale by exceeding the defined time limits or if the current
        day equals to `refresh_day_num`. Returns False if conditions are not met.

    Raises
    ------
    ValueError
        If the total period exceeds 7 days when `refresh_day_num` is set.
    ValidationError
        If `refresh_day_num` is not between 0 and 6.
    """

    table: Union[DeltaTableStep, str] = Field(
        ...,
        description="The table to check for stale data.",
    )
    interval: timedelta = Field(
        ...,
        description="The interval to consider data stale.",
    )
    refresh_day_num: Optional[int] = Field(
        default=None, description="The weekday number on which data should be refreshed.", ge=0, le=6
    )

    class Output(StepOutput):
        """Output class for StaleDataCheckStep."""

        is_data_stale: bool = Field(
            ..., description="Boolean flag indicating whether data in the table is stale or not"
        )

    @field_validator("table")
    def _validate_table(cls, table: Union[DeltaTableStep, str]) -> Union[DeltaTableStep, str]:
        """Validate `table` value"""
        if isinstance(table, str):
            return DeltaTableStep(table=table)
        return table

    @model_validator(mode="after")
    def _validate_refresh_day_num(self) -> "StaleDataCheckStep":
        """Validate input when `refresh_day_num` is provided."""
        if self.refresh_day_num is not None:
            max_period = timedelta(days=6, hours=23, minutes=59, seconds=59)
            if self.interval > max_period:
                raise ValueError("With refresh_day_num set, the total period must be less than 7 days.")

        return self

    def execute(self) -> Output:
        # Get the history of the Delta table
        history_df = self.table.describe_history()

        if not history_df:
            log.debug(f"No history found for `{self.table.table_name}`.")
            self.output.is_data_stale = True  # Consider data stale if the table does not exist
            return self.output

        modification_operations = [
            "WRITE",
            "MERGE",
            "DELETE",
            "UPDATE",
            "REPLACE TABLE AS SELECT",
            "CREATE TABLE AS SELECT",
            "TRUNCATE",
            "RESTORE",
        ]

        # Filter the history to data modification operations only
        history_df = history_df.filter(history_df["operation"].isin(modification_operations))

        # Get the last modification operation's timestamp
        last_modification = history_df.select("timestamp").first()

        if not last_modification:
            log.debug(f"No modification operation found in the history for `{self.table.table_name}`.")
            self.output.is_data_stale = True
            return self.output

        current_time = datetime.now()
        last_modification_timestamp = last_modification["timestamp"]

        cut_off_date = current_time - self.interval

        log.debug(f"Last modification timestamp: {last_modification_timestamp}, cut-off date: {cut_off_date}")

        is_stale_by_time = last_modification_timestamp <= cut_off_date

        if self.refresh_day_num is not None:
            current_day_of_week = current_time.weekday()
            log.debug(f"Current day of the week: {current_day_of_week}, refresh day: {self.refresh_day_num}")

            is_appropriate_day_for_refresh = current_day_of_week == self.refresh_day_num
            self.output.is_data_stale = is_stale_by_time or is_appropriate_day_for_refresh
            return self.output

        self.output.is_data_stale = is_stale_by_time



"""Kafka writer to write batch or streaming data into kafka topics"""

from typing import Dict, Optional, Union

from pyspark.sql import DataFrameWriter
from pyspark.sql.streaming import DataStreamWriter, StreamingQuery
from pyspark.sql.types import BinaryType, StringType

from koheesio.models import ExtraParamsMixin, Field, field_validator
from koheesio.spark.writers import Writer
from koheesio.spark.writers.stream import Trigger


class KafkaWriter(Writer, ExtraParamsMixin):
    """Kafka writer to write batch or streaming data into kafka topics

    All kafka specific options can be provided as additional init params

    Parameters
    ----------
    broker : str
        broker url of the kafka cluster
    topic : str
        full topic name to write the data to
    trigger : Optional[Union[Trigger, str, Dict]]
        Indicates optionally how to stream the data into kafka, continuous or batch
    checkpoint_location : str
        In case of a failure or intentional shutdown, you can recover the previous progress and state of a previous
        query, and continue where it left off. This is done using checkpointing and write-ahead logs.

    Example
    -------
    ```python
    KafkaWriter(
        write_broker="broker.com:9500",
        topic="test-topic",
        trigger=Trigger(continuous=True)
        includeHeaders: "true",
        key.serializer: "org.apache.kafka.common.serialization.StringSerializer",
        value.serializer: "org.apache.kafka.common.serialization.StringSerializer",
        kafka.group.id: "test-group",
        checkpoint_location: "s3://bucket/test-topic"
    )
    ```
    """

    format: str = "kafka"
    broker: str = Field(default=..., description="Kafka brokers to write to")
    topic: str = Field(default=..., description="Kafka topic to write to")
    trigger: Optional[Union[Trigger, str, Dict]] = Field(
        Trigger(available_now=True),
        description="Set the trigger for the stream query. If not set data is processed in batch",
    )
    checkpoint_location: str = Field(
        default=...,
        alias="checkpointLocation",
        description="In case of a failure or intentional shutdown, you can recover the previous progress and state of "
        "a previous query, and continue where it left off. This is done using checkpointing and write-ahead logs. You "
        "can configure a query with a checkpoint location, and the query will save all the progress information and "
        "the running aggregates to the checkpoint location. This checkpoint location has to be a path in an HDFS "
        "compatible file system that is accessible by Spark (e.g. Databricks Unity Catalog External Location)",
    )

    class Output(Writer.Output):
        """Output of the KafkaWriter"""

        streaming_query: Optional[Union[str, StreamingQuery]] = Field(
            default=None, description="Query ID of the stream query"
        )

    @property
    def streaming_query(self) -> Optional[Union[str, StreamingQuery]]:
        """return the streaming query"""
        return self.output.streaming_query

    @property
    def _trigger(self) -> dict[str, str]:
        """return the value of the Trigger object"""
        return self.trigger.value

    @field_validator("trigger")
    def _validate_trigger(cls, trigger: Optional[Union[Trigger, str, Dict]]) -> Trigger:
        """Validate the trigger value and convert it to a Trigger object if it is not already one."""
        return Trigger.from_any(trigger)

    def _validate_dataframe(self) -> None:
        """Validate the dataframe to be written to kafka"""
        if "value" not in self.df.columns:
            raise ValueError('Dataframe should at least have a "value" column.')
        if not isinstance(self.df.schema["value"].dataType, (StringType, BinaryType)):
            raise ValueError('The "value" column should be of type string or binary.')
        if "key" in self.df.columns:
            if not isinstance(self.df.schema["key"].dataType, (StringType, BinaryType)):
                raise ValueError('The "key" column should be of type string or binary.')

    @property
    def stream_writer(self) -> DataStreamWriter:
        """returns a stream writer

        Returns
        -------
        DataStreamWriter
        """
        write_stream = self.df.writeStream

        if self._trigger:
            write_stream = write_stream.trigger(**self._trigger)

        return write_stream

    @property
    def batch_writer(self) -> DataFrameWriter:
        """returns a batch writer

        Returns
        -------
        DataFrameWriter
        """
        return self.df.write

    @property
    def writer(self) -> Union[DataStreamWriter, DataFrameWriter]:
        """function to get the writer of proper type according to whether the data to written is a stream or not
        This function will also set the trigger property in case of a datastream.

        Returns
        -------
        Union[DataStreamWriter, DataFrameWriter]
            In case of streaming data -> DataStreamWriter, else -> DataFrameWriter
        """
        return self.stream_writer if self.streaming else self.batch_writer

    @property
    def options(self) -> Dict[str, str]:
        """retrieve the kafka options incl topic and broker.

        Returns
        -------
        dict
            Dict being the combination of kafka options + topic + broker
        """
        options = {
            **self.extra_params,
            "topic": self.topic,
            "kafka.bootstrap.servers": self.broker,
        }

        if self.checkpoint_location:
            options["checkpointLocation"] = self.checkpoint_location

        return options

    @property
    def logged_option_keys(self) -> set:
        """keys to be logged"""
        return {
            "kafka.bootstrap.servers",
            "topic",
            "includeHeaders",
            "key.serializer",
            "value.serializer",
            # "trigger",
            "checkpointLocation",
        }

    def execute(self) -> Writer.Output:
        """Effectively write the data from the dataframe (streaming of batch) to kafka topic.

        Returns
        -------
        KafkaWriter.Output
            streaming_query function can be used to gain insights on running write.
        """
        applied_options = {k: v for k, v in self.options.items() if k in self.logged_option_keys}
        self.log.debug(f"Applying options {applied_options}")

        self._validate_dataframe()

        _writer = self.writer.format(self.format).options(**self.options)
        self.output.streaming_query = _writer.start() if self.streaming else _writer.save()



"""
Module for the BaseReader class
"""

from typing import Optional, TypeVar
from abc import ABC, abstractmethod

from koheesio import Step

# Define a type variable that can be any type of DataFrame
DataFrameType = TypeVar("DataFrameType")


class BaseReader(Step, ABC):
    """Base class for all Readers

    A Reader is a Step that reads data from a source based on the input parameters
    and stores the result in self.output.df (DataFrame).

    When implementing a Reader, the execute() method should be implemented.
    The execute() method should read from the source and store the result in self.output.df.

    The Reader class implements a standard read() method that calls the execute() method and returns the result. This
    method can be used to read data from a Reader without having to call the execute() method directly. Read method
    does not need to be implemented in the child class.

    The Reader class also implements a shorthand for accessing the output Dataframe through the df-property. If the
    output.df is None, .execute() will be run first.
    """

    @property
    def df(self) -> Optional[DataFrameType]:
        """Shorthand for accessing self.output.df
        If the output.df is None, .execute() will be run first
        """
        if not self.output.df:
            self.execute()
        return self.output.df

    @abstractmethod
    def execute(self) -> Step.Output:
        """Execute on a Reader should handle self.output.df (output) as a minimum
        Read from whichever source -> store result in self.output.df
        """
        pass

    def read(self) -> DataFrameType:
        """Read from a Reader without having to call the execute() method directly"""
        self.execute()
        return self.output.df



"""
Create Spark DataFrame from table in Metastore
"""

from koheesio.models import Field
from koheesio.spark.readers import Reader


class MetastoreReader(Reader):
    """Reader for tables/views from Spark Metastore

    Parameters
    ----------
    table : str
        Table name in spark metastore
    """

    table: str = Field(default=..., description="Table name in spark metastore")

    def execute(self) -> Reader.Output:
        self.output.df = self.spark.table(self.table)



from unittest.mock import patch

from conftest import ScopeSecrets

from koheesio.integrations.spark.databricks.secrets import DataBricksSecret


class TestDatabricksSecret:
    def test_set_parent_to_scope(self):
        # Test when parent is not provided
        secret = DataBricksSecret(scope="secret-scope")
        assert secret.parent == "secret_scope"

        # Test when parent is provided
        secret = DataBricksSecret(scope="secret-scope", parent="custom_parent")
        assert secret.parent == "custom_parent"

    @patch("koheesio.integrations.spark.databricks.secrets.DataBricksSecret._client")
    def test_get_secrets_no_alias(self, mock_databricks_client):
        with patch("koheesio.integrations.spark.databricks.utils.on_databricks", return_value=True):
            dd = {
                "key1": "value_of_key1",
                "key2": "value_of_key2",
            }
            databricks = DataBricksSecret(scope="dummy", parent="kafka")
            mock_databricks_client.secrets = ScopeSecrets(dd)
            secrets = databricks._get_secrets()

            assert secrets["key1"] == "value_of_key1"
            assert secrets["key2"] == "value_of_key2"

    @patch("koheesio.integrations.spark.databricks.secrets.DataBricksSecret._client")
    def test_get_secrets_alias(self, mock_databricks_client):
        with patch("koheesio.integrations.spark.databricks.utils.on_databricks", return_value=True):
            dd = {
                "key1": "value_of_key1",
                "key2": "value_of_key2",
            }
            alias = {
                "key1": "new_name_key1",
                "key2": "new_name_key2",
            }
            databricks = DataBricksSecret(scope="dummy", parent="kafka", alias=alias)
            mock_databricks_client.secrets = ScopeSecrets(dd)
            secrets = databricks._get_secrets()

            assert secrets["new_name_key1"] == "value_of_key1"
            assert secrets["new_name_key2"] == "value_of_key2"



"""
Spark Utility functions
"""

from typing import Union
from enum import Enum
import importlib
import inspect
import os
from types import ModuleType
import warnings

from pyspark import sql
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    NullType,
    ShortType,
    StringType,
    StructType,
    TimestampType,
)
from pyspark.version import __version__ as spark_version

__all__ = [
    "SparkDatatype",
    "import_pandas_based_on_pyspark_version",
    "on_databricks",
    "schema_struct_to_schema_str",
    "spark_data_type_is_array",
    "spark_data_type_is_numeric",
    "show_string",
    "get_spark_minor_version",
    "SPARK_MINOR_VERSION",
    "AnalysisException",
    "Column",
    "DataFrame",
    "SparkSession",
    "ParseException",
    "DataType",
    "DataFrameReader",
    "DataStreamReader",
    "DataFrameWriter",
    "DataStreamWriter",
    "StreamingQuery",
    "Window",
    "WindowSpec",
    "get_active_session",
    "check_if_pyspark_connect_module_is_available",
    "check_if_pyspark_connect_is_supported",
    "get_column_name",
]

try:
    from pyspark.errors.exceptions.base import AnalysisException  # type: ignore
except (ImportError, ModuleNotFoundError):
    from pyspark.sql.utils import AnalysisException  # type: ignore


AnalysisException = AnalysisException


def get_spark_minor_version() -> float:
    """Returns the minor version of the spark instance.

    For example, if the spark version is 3.3.2, this function would return 3.3
    """
    return float(".".join(spark_version.split(".")[:2]))


# shorthand for the get_spark_minor_version function
SPARK_MINOR_VERSION: float = get_spark_minor_version()


class PysparkConnectModuleNotAvailableWarning(Warning):
    """Warning to be raised when the pyspark connect module is not available"""

    def __init__(self):
        message = (
            "It looks like the required modules for Spark Connect (e.g. grpcio) are not installed. "
            "If not, you can install them using `pip install pyspark[connect]` or `koheesio[pyspark_connect]`. "
        )
        super().__init__(message)


def check_if_pyspark_connect_module_is_available() -> bool:
    """Check if the pyspark connect module is available

    Returns
    -------
    bool
        True if the pyspark connect module is available, False otherwise.

    Raises
    ------
    ImportError
        If the required modules for Spark Connect are not importable.
    """

    # before pyspark 3.4, connect was not supported
    if SPARK_MINOR_VERSION <= 3.4:
        return False

    try:
        importlib.import_module("pyspark.sql.connect")
        # check extras: grpcio package is needed for pyspark[connect] to work
        try:
            importlib.import_module("grpc")
        except (ImportError, ModuleNotFoundError) as e_import:
            raise e_import
        return True
    except (ImportError, ModuleNotFoundError):
        warnings.warn(PysparkConnectModuleNotAvailableWarning())
        return False


def check_if_pyspark_connect_is_supported() -> bool:
    warnings.warn(
        message=(
            "The `check_if_pyspark_connect_is_supported` function has been"
            " replaced by `check_if_pyspark_connect_module_is_available`."
            " Import it instead. Current function will be removed in the future."
        ),
        category=DeprecationWarning,
        stacklevel=2,
    )
    return check_if_pyspark_connect_module_is_available()


if check_if_pyspark_connect_module_is_available():
    """Only import the connect module if the current version of PySpark supports it"""
    from pyspark.errors.exceptions.captured import (
        ParseException as CapturedParseException,
    )
    from pyspark.errors.exceptions.connect import (
        ParseException as ConnectParseException,
    )
    from pyspark.sql.connect.column import Column as ConnectColumn
    from pyspark.sql.connect.dataframe import DataFrame as ConnectDataFrame
    from pyspark.sql.connect.proto.types_pb2 import DataType as ConnectDataType
    from pyspark.sql.connect.readwriter import DataFrameReader, DataFrameWriter
    from pyspark.sql.connect.session import SparkSession as ConnectSparkSession
    from pyspark.sql.connect.streaming.readwriter import (
        DataStreamReader,
        DataStreamWriter,
    )
    from pyspark.sql.connect.window import Window as ConnectWindow
    from pyspark.sql.connect.window import WindowSpec as ConnectWindowSpec
    from pyspark.sql.streaming.query import StreamingQuery
    from pyspark.sql.types import DataType as SqlDataType
    from pyspark.sql.window import Window, WindowSpec

    Column = Union[sql.Column, ConnectColumn]
    DataFrame = Union[sql.DataFrame, ConnectDataFrame]
    SparkSession = Union[sql.SparkSession, ConnectSparkSession]
    ParseException = (CapturedParseException, ConnectParseException)
    DataType = Union[SqlDataType, ConnectDataType]
    DataFrameReader = Union[sql.readwriter.DataFrameReader, DataFrameReader]  # type: ignore
    DataStreamReader = Union[sql.streaming.readwriter.DataStreamReader, DataStreamReader]  # type: ignore
    DataFrameWriter = Union[sql.readwriter.DataFrameWriter, DataFrameWriter]  # type: ignore
    DataStreamWriter = Union[sql.streaming.readwriter.DataStreamWriter, DataStreamWriter]  # type: ignore
    StreamingQuery = StreamingQuery
    Window = Union[Window, ConnectWindow]
    WindowSpec = Union[WindowSpec, ConnectWindowSpec]
else:
    """Import the regular PySpark modules if the current version of PySpark does not support the connect module"""
    try:
        from pyspark.errors.exceptions.captured import ParseException  # type: ignore
    except (ImportError, ModuleNotFoundError):
        from pyspark.sql.utils import ParseException  # type: ignore

    ParseException = ParseException

    from pyspark.sql.column import Column  # type: ignore
    from pyspark.sql.dataframe import DataFrame  # type: ignore
    from pyspark.sql.readwriter import DataFrameReader, DataFrameWriter  # type: ignore
    from pyspark.sql.session import SparkSession  # type: ignore
    from pyspark.sql.types import DataType  # type: ignore
    from pyspark.sql.window import Window, WindowSpec  # type: ignore

    try:
        from pyspark.sql.streaming.query import StreamingQuery  # type: ignore
        from pyspark.sql.streaming.readwriter import DataStreamReader, DataStreamWriter  # type: ignore
    except (ImportError, ModuleNotFoundError):
        from pyspark.sql.streaming import (  # type: ignore
            DataStreamReader,
            DataStreamWriter,
            StreamingQuery,
        )
    DataFrameReader = DataFrameReader
    DataStreamReader = DataStreamReader
    DataFrameWriter = DataFrameWriter
    DataStreamWriter = DataStreamWriter
    StreamingQuery = StreamingQuery
    Window = Window
    WindowSpec = WindowSpec


def get_active_session() -> SparkSession:  # type: ignore
    """Get the active Spark session"""
    if check_if_pyspark_connect_module_is_available() and SPARK_MINOR_VERSION >= 3.5:
        from pyspark.sql.connect.session import SparkSession as _ConnectSparkSession

        session = _ConnectSparkSession.getActiveSession() or sql.SparkSession.getActiveSession()  # type: ignore

    else:
        session = sql.SparkSession.getActiveSession()

    if not session:
        raise RuntimeError(
            "No active Spark session found. Please create a Spark session before using module connect_utils."
            " Or perform local import of the module."
        )

    return session


class SparkDatatype(Enum):
    """
    Allowed spark datatypes

    The following table lists the data types that are supported by Spark SQL.

    | Data type	    | SQL name                  |
    |---------------|---------------------------|
    | ByteType      | BYTE, TINYINT             |
    | ShortType     | SHORT, SMALLINT           |
    | IntegerType   | INT, INTEGER              |
    | LongType      | LONG, BIGINT              |
    | FloatType     | FLOAT, REAL               |
    | DoubleType    | DOUBLE                    |
    | DecimalType   | DECIMAL, DEC, NUMERIC     |
    | StringType    | STRING                    |
    | BinaryType    | BINARY                    |
    | BooleanType	| BOOLEAN                   |
    | TimestampType | TIMESTAMP, TIMESTAMP_LTZ  |
    | DateType      | DATE                      |
    | ArrayType     | ARRAY<element_type>       |
    | MapType       | MAP<key_type, value_type> |
    | NullType      | VOID                      |

    Not supported yet
    ----------------
    * __TimestampNTZType__
        TIMESTAMP_NTZ
    * __YearMonthIntervalType__
        INTERVAL YEAR, INTERVAL YEAR TO MONTH, INTERVAL MONTH
    * __DayTimeIntervalType__
        INTERVAL DAY, INTERVAL DAY TO HOUR, INTERVAL DAY TO MINUTE, INTERVAL DAY TO SECOND, INTERVAL HOUR,
        INTERVAL HOUR TO MINUTE, INTERVAL HOUR TO SECOND, INTERVAL MINUTE, INTERVAL MINUTE TO SECOND, INTERVAL SECOND

    See Also
    --------
    https://spark.apache.org/docs/latest/sql-ref-datatypes.html#supported-data-types
    """

    # byte
    BYTE = "byte"
    TINYINT = "byte"

    # short
    SHORT = "short"
    SMALLINT = "short"

    # integer
    INTEGER = "integer"
    INT = "integer"

    # long
    LONG = "long"
    BIGINT = "long"

    # float
    FLOAT = "float"
    REAL = "float"

    # timestamp
    TIMESTAMP = "timestamp"
    TIMESTAMP_LTZ = "timestamp"

    # decimal
    DECIMAL = "decimal"
    DEC = "decimal"
    NUMERIC = "decimal"

    DATE = "date"
    DOUBLE = "double"
    STRING = "string"
    BINARY = "binary"
    BOOLEAN = "boolean"
    ARRAY = "array"
    MAP = "map"
    VOID = "void"

    @property
    def spark_type(self) -> type:
        """Returns the spark type for the given enum value"""
        mapping_dict = {
            "byte": ByteType,
            "short": ShortType,
            "integer": IntegerType,
            "long": LongType,
            "float": FloatType,
            "double": DoubleType,
            "decimal": DecimalType,
            "string": StringType,
            "binary": BinaryType,
            "boolean": BooleanType,
            "timestamp": TimestampType,
            "date": DateType,
            "array": ArrayType,
            "map": MapType,
            "void": NullType,
        }
        return mapping_dict[self.value]

    @classmethod
    def from_string(cls, value: str) -> "SparkDatatype":
        """Allows for getting the right Enum value by simply passing a string value
        This method is not case-sensitive
        """
        return getattr(cls, value.upper())


def on_databricks() -> bool:
    """Retrieve if we're running on databricks or elsewhere"""
    dbr_version = os.getenv("DATABRICKS_RUNTIME_VERSION", None)
    spark = get_active_session()
    dbr_spark_version = spark.conf.get("spark.databricks.clusterUsageTags.effectiveSparkVersion", None)
    return (dbr_version is not None and dbr_version != "") or (dbr_spark_version is not None)


def spark_data_type_is_array(data_type: DataType) -> bool:  # type: ignore
    """Check if the column's dataType is of type ArrayType"""
    return isinstance(data_type, ArrayType)


def spark_data_type_is_numeric(data_type: DataType) -> bool:  # type: ignore
    """Check if the column's dataType is of type ArrayType"""
    return isinstance(data_type, (IntegerType, LongType, FloatType, DoubleType, DecimalType))


def schema_struct_to_schema_str(schema: StructType) -> str:
    """Converts a StructType to a schema str"""
    if not schema:
        return ""
    return ",\n".join([f"{field.name} {field.dataType.typeName().upper()}" for field in schema.fields])


def import_pandas_based_on_pyspark_version() -> ModuleType:
    """
    This function checks the installed version of PySpark and then tries to import the appropriate version of pandas.
    If the correct version of pandas is not installed, it raises an ImportError with a message indicating which version
    of pandas should be installed.
    """
    try:
        import pandas as pd

        pyspark_version = get_spark_minor_version()
        pandas_version = pd.__version__

        if (pyspark_version < 3.4 and pandas_version >= "2") or (pyspark_version >= 3.4 and pandas_version < "2"):
            raise ImportError(
                f"For PySpark {pyspark_version}, "
                f"please install Pandas version {'< 2' if pyspark_version < 3.4 else '>= 2'}"
            )

        return pd
    except ImportError as e:
        raise ImportError("Pandas module is not installed.") from e


def show_string(df: DataFrame, n: int = 20, truncate: Union[bool, int] = True, vertical: bool = False) -> str:  # type: ignore
    """Returns a string representation of the DataFrame
    The default implementation of DataFrame.show() hardcodes a print statement, which is not always desirable.
    With this function, you can get the string representation of the DataFrame instead, and choose how to display it.

    Example
    -------
    ```python
    print(show_string(df))

    # or use with a logger
    logger.info(show_string(df))
    ```

    Parameters
    ----------
    df : DataFrame
        The DataFrame to display
    n : int, optional
        The number of rows to display, by default 20
    truncate : Union[bool, int], optional
        If set to True, truncate the displayed columns, by default True
    vertical : bool, optional
        If set to True, display the DataFrame vertically, by default False
    """
    if SPARK_MINOR_VERSION < 3.5:
        # noinspection PyProtectedMember
        return df._jdf.showString(n, truncate, vertical)  # type: ignore
    # as per spark 3.5, the _show_string method is now available making calls to _jdf.showString obsolete
    # noinspection PyProtectedMember
    return df._show_string(n, truncate, vertical)


def get_column_name(col: Column) -> str:  # type: ignore
    """Get the column name from a Column object

    Normally, the name of a Column object is not directly accessible in the regular pyspark API. This function
    extracts the name of the given column object without needing to provide it in the context of a DataFrame.

    Parameters
    ----------
    col: Column
        The Column object

    Returns
    -------
    str
        The name of the given column
    """
    # we have to distinguish between the Column object from column from local session and remote
    if hasattr(col, "_jc"):
        # In case of a 'regular' Column object, we can directly access the name attribute through the _jc attribute
        # noinspection PyProtectedMember
        name = col._jc.toString()  # type: ignore[operator]
    elif any(cls.__module__ == "pyspark.sql.connect.column" for cls in inspect.getmro(col.__class__)):
        # noinspection PyProtectedMember
        name = col._expr.name()
    else:
        raise ValueError("Column object is not a valid Column object")

    return name



import pytest

from pyspark.sql import Column
from pyspark.sql import functions as f

from koheesio.spark.transformations import (
    ColumnsTransformation,
    ColumnsTransformationWithTarget,
)
from koheesio.spark.transformations.dummy import DummyTransformation
from koheesio.spark.utils import SparkDatatype

pytestmark = pytest.mark.spark


def test_transform(dummy_df):
    tf = DummyTransformation()
    df = tf.transform(dummy_df)
    actual = df.head().asDict()
    expected = {"id": 0, "hello": "world"}
    assert actual == expected
    assert df == tf.output.df


def test_transform_unhappy():
    """Calling transform without having set or passed a DataFrame should raise a RuntimeError"""
    tf = DummyTransformation()
    with pytest.raises(RuntimeError):
        tf.transform()


def test_execute(dummy_df):
    tf = DummyTransformation(df=dummy_df)
    output = tf.execute()
    df = output.df
    actual = df.head().asDict()
    expected = {"id": 0, "hello": "world"}
    assert actual == expected
    assert df == tf.output.df


class TestColumnsTransformation:
    data = [["a", 1], ["b", 2]]
    columns = ["str_column", "long_column"]

    @pytest.fixture(scope="class")
    def input_df(self, spark):
        df = spark.createDataFrame(self.data, self.columns)
        yield df

    class AddOne(ColumnsTransformation):
        def execute(self):
            for column in self.get_columns():
                self.output.df = self.df.withColumn(column, f.col(column) + 1)

    @pytest.fixture(scope="class")
    def add_one(self, input_df):
        yield self.AddOne(columns=self.columns, df=input_df)

    def test_columns_transformation(self, add_one):
        expected = {"str_column": "a", "long_column": 2}
        output_df = add_one.execute().df
        actual = output_df.head().asDict()
        assert actual == expected

    def test_column_type_of_col(self, add_one):
        assert add_one.column_type_of_col(f.col("str_column")) == "string"
        assert add_one.column_type_of_col(f.col("long_column")) == "long"

    class AddOneWithTarget(ColumnsTransformationWithTarget):
        _limit_data_type = "long"

        def func(self, column: Column):
            return column + 1

    @pytest.fixture()
    def add_one_with_target(self, input_df):
        yield self.AddOneWithTarget(columns=["long_column"], df=input_df, target_column="target_column")

    def test_columns_transformation_with_target(self, add_one_with_target):
        expected = {"long_column": 1, "str_column": "a", "target_column": 2}
        output_df = add_one_with_target.execute().df
        actual = output_df.head().asDict()
        assert actual == expected


class TestColumnsTransformationDataTypeLimitations:
    data = [["a", 1], ["b", 2]]
    columns = ["str_column", "long_column"]

    @pytest.fixture(scope="class")
    def str_long_df(self, spark):
        df = spark.createDataFrame(self.data, self.columns)
        yield df

    class AddTwoStrict(ColumnsTransformationWithTarget):
        class ColumnConfig:
            limit_data_type = [SparkDatatype.LONG]
            data_type_strict_mode = True
            run_for_all_data_type = [SparkDatatype.LONG]

        def func(self, column: Column):
            return column + 2

    def test_strict_config_with_columns_happy(self, str_long_df):
        # explicitly pass a value for columns
        add_two = self.AddTwoStrict(
            columns=["long_column"],
            df=str_long_df,
            target_column="target_column",
        )

        # we should get None on run_for_all_is_set since we have specified columns
        assert add_two.run_for_all_is_set is True
        assert add_two.ColumnConfig.run_for_all_data_type == [SparkDatatype.LONG]

        assert add_two.limit_data_type_is_set is True
        assert add_two.ColumnConfig.limit_data_type == [SparkDatatype.LONG]

        assert add_two.data_type_strict_mode_is_set is True
        assert add_two.ColumnConfig.data_type_strict_mode is True

        columns = [col for col in add_two.get_columns()]
        assert columns == ["long_column"]

    def test_strict_config_with_columns_unhappy(self, str_long_df):
        """Passing a column with a 'wrong' datatype should raise a ValueError when strict mode is enabled"""
        # explicitly pass a value for columns
        add_two = self.AddTwoStrict(
            columns=["str_column"],
            df=str_long_df,
            target_column="target_column",
        )

        assert add_two.data_type_strict_mode_is_set is True
        assert add_two.ColumnConfig.data_type_strict_mode is True

        assert add_two.limit_data_type_is_set is True
        assert add_two.ColumnConfig.limit_data_type == [SparkDatatype.LONG]

        with pytest.raises(ValueError):
            _ = [col for col in add_two.get_columns()]

    def test_strict_mode(self, str_long_df):
        add_two = self.AddTwoStrict(
            columns=["str_column"],
            df=str_long_df,
        )
        with pytest.raises(ValueError):
            add_two.execute()

    class AddTwoLongOnly(AddTwoStrict):
        class ColumnConfig:
            limit_data_type = [None]
            data_type_strict_mode = False
            run_for_all_data_type = [SparkDatatype.LONG]

    def test_run_all_without_columns_happy(self, str_long_df):
        add_two = self.AddTwoLongOnly(
            df=str_long_df,
            target_suffix="_add_two",
        )

        assert add_two.data_type_strict_mode_is_set is False
        assert add_two.ColumnConfig.data_type_strict_mode is False

        assert add_two.limit_data_type_is_set is False
        assert add_two.ColumnConfig.limit_data_type == [None]

        assert add_two.run_for_all_is_set is True
        assert add_two.ColumnConfig.run_for_all_data_type == [SparkDatatype.LONG]

        columns = [col for col in add_two.get_columns()]
        assert columns == ["long_column"]

    class MultipleDataTypes(ColumnsTransformationWithTarget):
        class ColumnConfig:
            limit_data_type = [SparkDatatype.LONG, SparkDatatype.STRING]
            data_type_strict_mode = True
            run_for_all_data_type = [SparkDatatype.LONG, SparkDatatype.STRING]

        def func(self, column: Column):
            return column

    def test_multiple_datatypes_happy(self, str_long_df):
        multiple_data_types = self.MultipleDataTypes(
            df=str_long_df,
        )

        assert multiple_data_types.data_type_strict_mode_is_set is True
        assert multiple_data_types.ColumnConfig.data_type_strict_mode is True

        assert multiple_data_types.limit_data_type_is_set is True
        assert multiple_data_types.ColumnConfig.limit_data_type == [SparkDatatype.LONG, SparkDatatype.STRING]

        assert multiple_data_types.run_for_all_is_set is True
        assert multiple_data_types.ColumnConfig.run_for_all_data_type == [SparkDatatype.LONG, SparkDatatype.STRING]

        actual = [col for col in multiple_data_types.get_columns()]
        expected = ["str_column", "long_column"]
        assert sorted(actual) == sorted(expected)

    def test_column_does_not_exist(self, str_long_df):
        class TestTransformation(ColumnsTransformation):
            def execute(self):
                pass

        tf = TestTransformation(columns=["non_existent_column"], df=str_long_df)

        with pytest.raises(ValueError, match="Column 'non_existent_column' does not exist in the DataFrame schema"):
            tf.column_type_of_col("non_existent_column")



"""
Module for KafkaReader and KafkaStreamReader.
"""

from typing import Dict, Optional

from koheesio.models import ExtraParamsMixin, Field
from koheesio.spark import DataFrameReader, DataStreamReader
from koheesio.spark.readers import Reader


class KafkaReader(Reader, ExtraParamsMixin):
    """
    Reader for Kafka topics.

    Wrapper around Spark's kafka read format. Supports both batch and streaming reads.

    Parameters
    ----------
    read_broker : str
        Kafka brokers to read from. Should be passed as a single string with multiple brokers passed in a comma
        separated list
    topic : str
        Kafka topic to consume.
    streaming : Optional[bool]
        Whether to read the kafka topic as a stream or not.
    params : Optional[Dict[str, str]]
        Arbitrary options to be applied when creating NSP Reader. If a user provides values for `subscribe` or
        `kafka.bootstrap.servers`, they will be ignored in favor of configuration passed through `topic` and
        `read_broker` respectively. Defaults to an empty dictionary.

    Notes
    -----
    * The `read_broker` and `topic` parameters are required.
    * The `streaming` parameter defaults to `False`.
    * The `params` parameter defaults to an empty dictionary. This parameter is also aliased as `kafka_options`.
    * Any extra kafka options can also be passed as key-word arguments; these will be merged with the `params` parameter

    Example
    -------
    ```python
    from koheesio.spark.readers.kafka import KafkaReader

    kafka_reader = KafkaReader(
        read_broker="kafka-broker-1:9092,kafka-broker-2:9092",
        topic="my-topic",
        streaming=True,
        # extra kafka options can be passed as key-word arguments
        startingOffsets="earliest",
    )
    ```

    In the example above, the `KafkaReader` will read from the `my-topic` Kafka topic, using the brokers
    `kafka-broker-1:9092` and `kafka-broker-2:9092`. The reader will read the topic as a stream and will start reading
    from the earliest available offset.

    The stream can be started by calling the `read` or `execute` method on the `kafka_reader` object.

    > Note: The `KafkaStreamReader` could be used in the example above to achieve the same result. `streaming` would
        default to `True` in that case and could be omitted from the parameters.

    See Also
    --------
    - Official Spark Documentation: https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html
    """

    read_broker: str = Field(
        ...,
        description="Kafka brokers to read from, should be passed as a single string with multiple brokers passed in a "
        "comma separated list",
    )

    topic: str = Field(default=..., description="Kafka topic to consume.")
    streaming: Optional[bool] = Field(
        default=False, description="Whether to read the kafka topic as a stream or not. Defaults to False."
    )
    params: Dict[str, str] = Field(
        default_factory=dict,
        alias="kafka_options",
        description="Arbitrary options to be applied when creating NSP Reader. If a user provides values for "
        "'subscribe' or 'kafka.bootstrap.servers', they will be ignored in favor of configuration passed through "
        "'topic' and 'read_broker' respectively.",
    )

    @property
    def stream_reader(self) -> DataStreamReader:
        """Returns the Spark readStream object."""
        return self.spark.readStream

    @property
    def batch_reader(self) -> DataFrameReader:
        """Returns the Spark read object for batch processing."""
        return self.spark.read

    @property
    def reader(self) -> Reader:
        """Returns the appropriate reader based on the streaming flag."""
        if self.streaming:
            return self.stream_reader
        return self.batch_reader

    @property
    def options(self) -> Dict[str, str]:
        """Merge fixed parameters with arbitrary options provided by user."""
        return {
            **self.params,
            "subscribe": self.topic,
            "kafka.bootstrap.servers": self.read_broker,
        }

    @property
    def logged_option_keys(self) -> set:
        """Keys that are allowed to be logged for the options."""
        return {
            "kafka.bootstrap.servers",
            "subscribe",
            "subscribePattern",
            "includeHeaders",
            "startingOffsets",
            "endingOffsets",
            "key.deserializer",
            "value.deserializer",
            "kafka.group.id",
        }

    def execute(self) -> Reader.Output:
        applied_options = {k: v for k, v in self.options.items() if k in self.logged_option_keys}
        self.log.debug(f"Applying options {applied_options}")

        self.output.df = self.reader.format("kafka").options(**self.options).load()  # type: ignore


class KafkaStreamReader(KafkaReader):
    """KafkaStreamReader is a KafkaReader that reads data as a stream

    This class is identical to KafkaReader, with the `streaming` parameter defaulting to `True`.
    """

    streaming: bool = True



from pathlib import Path
from textwrap import dedent

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.sql_transform import SqlTransform

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_sql_transform")


@pytest.fixture
def test_data_path(data_path) -> Path:
    return Path(data_path) / "transformations"


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # Description: simple sql script
            # input values
            dict(table_name="dummy", sql="SELECT id AS new_id FROM ${table_name}"),
            # expected output
            {"new_id": 0},
        ),
        (
            # Description: Using explicit params (new_column)
            # input values
            dict(
                table_name="dummy_table",
                params=dict(new_column="incr_id"),  # explicit param
                sql="SELECT id, id + 1 as ${new_column} FROM ${table_name}",
            ),
            # expected output
            {"id": 0, "incr_id": 1},
        ),
        (
            # Description: Using implicit params (new_column)
            # input values
            dict(
                table_name="dummy_table",
                new_column="incr_id",  # implicit param
                sql="SELECT id, id + 1 as ${new_column} FROM ${table_name}",
            ),
            # expected output
            {"id": 0, "incr_id": 1},
        ),
        (
            # Description: with sql_path as Path like object
            # input values
            dict(
                table_name="dummy_table",
                sql_path="dummy.sql",
            ),
            # expected output
            {"id": 0, "incremented_id": 1},
        ),
        (
            # Description: with sql_path as str
            # input values
            dict(
                table_name="dummy_table",
                sql_path="dummy.sql",
            ),
            # expected output
            {"id": 0, "incremented_id": 1},
        ),
    ],
)
def test_sql_transform(input_values, expected, dummy_df, test_data_path):
    if sql_path := input_values.get("sql_path"):
        input_values["sql_path"] = str((test_data_path / sql_path).as_posix())
    result = SqlTransform(**input_values).transform(dummy_df)
    actual = result.head().asDict()

    log.info(
        dedent(
            f"""
            input values: {input_values}
            actual: {actual}
            expected: {expected}
            """
        )
    )

    assert expected == actual


@pytest.mark.parametrize(
    "input_values,error",
    [
        (
            # non-existent path
            dict(table_name="...", sql_path="../none_existent_path"),
            FileNotFoundError,
        ),
        (
            # both sql and sql_path specified
            dict(table_name="...", sql_path="foo", sql="bar"),
            ValueError,
        ),
        (
            # no sql or sql_path specified
            dict(table_name="..."),
            ValueError,
        ),
        (
            # missing required keyword-only argument
            dict(),
            ValueError,
        ),
    ],
)
def test_errors(input_values, error):
    log.info(
        dedent(
            f"""
            input values: {input_values}
            expected error: {error}
            """
        )
    )

    with pytest.raises(error):
        SqlTransform(**input_values)



"""
Test the UpperCase columns transformation
"""

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.change_case import (
    InitCap,
    LowerCase,
    TitleCase,
    UpperCase,
)
from koheesio.spark.utils import show_string

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)
data = [
    ["Banana lemon orange", 1000, "USA"],
    ["Carrots blueberries", 1500, None],
    ["Beans", 1600, "USA"],
]
columns = ["product", "amount", "country"]


@pytest.mark.parametrize(
    "input_values,input_data,input_schema,expected",
    [
        (
            # description: Test a normal call to LowerCase, UpperCase, TitleCase, InitCap
            # input values
            dict(
                column="product",
                target_column="result",
            ),
            # input data and schema
            data,
            columns,
            # expected data
            dict(
                LowerCase=["banana lemon orange", "carrots blueberries", "beans"],
                UpperCase=["BANANA LEMON ORANGE", "CARROTS BLUEBERRIES", "BEANS"],
                TitleCase=["Banana Lemon Orange", "Carrots Blueberries", "Beans"],
                InitCap=["Banana Lemon Orange", "Carrots Blueberries", "Beans"],
            ),
        ),
        (
            # description: Test a call with null values to LowerCase, UpperCase, TitleCase, InitCap
            # input values
            dict(
                columns=["country"],
                target_column="result",
            ),
            # input data and schema
            data,
            columns,
            # expected data
            dict(
                LowerCase=["usa", None, "usa"],
                UpperCase=["USA", None, "USA"],
                TitleCase=["Usa", None, "Usa"],
                InitCap=["Usa", None, "Usa"],
            ),
        ),
    ],
)
def test_happy_flow(input_values, input_data, input_schema, expected, spark):
    input_df = spark.createDataFrame(input_data, input_schema)

    klz = LowerCase, UpperCase, TitleCase, InitCap

    for kls in klz:
        log.info(f"Testing {kls.__name__}")
        change_case = kls(**input_values)
        output_df = change_case.transform(input_df)
        target_column = change_case.target_column

        # log equivalent of doing df.show()
        log.info(f"show output_df: \n{show_string(output_df, 20, 20, False)}")

        actual = [row[target_column] for row in output_df.select(target_column).collect()]
        assert actual == expected[kls.__name__]


# TODO: not supported yet
#
#     def test_transform_nested_df(self):  # LowerCase
#         """
#         Test a nested dataframe call
#         """
#         output_df = (
#             LowerCase(
#                 df=self.sample_nested_df,
#                 params={"source_column": "payload.col2.col3"},
#                 target_alias="case_converted",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         output_list = output_df.select("case_converted").rdd.map(lambda r: r[0]).collect()
#
#         self.assertEqual(output_list, ["value1"])


# TODO: not supported yet
#
#     def test_transform_nested_df(self):  # UpperCase
#         """
#         Test a nested dataframe call
#         """
#         output_df = (
#             UpperCase(
#                 df=self.sample_nested_df,
#                 params={"source_column": "payload.col2.col3"},
#                 target_alias="case_converted",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         output_list = output_df.select("case_converted").rdd.map(lambda r: r[0]).collect()
#
#         self.assertEqual(output_list, ["VALUE1"])


#   TODO: not supported yet
#
#     def test_transform_nested_df(self):  # TitleCase
#         """
#         Test a nested dataframe call
#         """
#         output_df = (
#             TitleCase(
#                 df=self.sample_nested_df,
#                 params={"source_column": "payload.col2.col3"},
#                 target_alias="case_converted",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         output_list = output_df.select("case_converted").rdd.map(lambda r: r[0]).collect()
#
#         self.assertEqual(output_list, ["Value1"])



from aiohttp import ClientSession, TCPConnector
from aiohttp_retry import ExponentialRetry
from aioresponses import aioresponses
import pytest
import responses
from responses.registries import OrderedRegistry
from yarl import URL

from pyspark.sql.types import MapType, StringType, StructField, StructType

from koheesio.asyncio.http import AsyncHttpStep
from koheesio.spark.readers.rest_api import AsyncHttpGetStep, RestApiReader
from koheesio.steps.http import HttpGetStep, PaginatedHttpGetStep

ASYNC_BASE_URL = "https://42.koheesio.test"
ASYNC_GET_ENDPOINT = URL(f"{ASYNC_BASE_URL}/get")

pytestmark = pytest.mark.spark


@pytest.fixture(scope="function", name="mock_aiohttp")
def mock_aiohttp():
    with aioresponses() as m:
        yield m


@responses.activate(registry=OrderedRegistry)
def test_paginated_api():
    for i in range(1, 4):  # Mock 3 pages of data
        data = [{"id": j, "page": i, "value": f"data_{i}_{j}"} for j in range(1, 11)]  # 10 records per page
        responses.get(f"https://api.example.com/data?page={i}", json=data)

    # Test that the paginated API returns all the data
    transport = PaginatedHttpGetStep(url="https://api.example.com/data?page={page}", paginate=True, pages=3)
    task = RestApiReader(transport=transport, spark_schema="id: int, page:int, value: string")

    assert isinstance(task.transport, PaginatedHttpGetStep)

    task.execute()

    # Convert the DataFrame to a list of dictionaries
    all_data = [row.asDict() for row in task.output.df.collect()]
    expected_data = [{"id": j, "page": i, "value": f"data_{i}_{j}"} for i in range(1, 4) for j in range(1, 11)]

    assert all_data == expected_data


@pytest.mark.asyncio
async def test_async_rest_api_reader(mock_aiohttp):
    """
    Testing the AsyncHttpStep class.
    """
    mock_aiohttp.get(str(ASYNC_GET_ENDPOINT), status=200, repeat=True, payload={"url": str(ASYNC_GET_ENDPOINT)})

    transport = AsyncHttpGetStep(
        client_session=ClientSession(),
        url=[URL(ASYNC_GET_ENDPOINT), URL(ASYNC_GET_ENDPOINT)],
        retry_options=ExponentialRetry(),
        connector=TCPConnector(limit=10),
        headers={"Content-Type": "application/json", "X-type": "Koheesio RestApiReader Test"},
    )

    spark_schema = StructType(
        [
            StructField("origin", StringType(), True),
            StructField("url", StringType(), True),
        ]
    )
    task = RestApiReader(transport=transport, spark_schema=spark_schema)

    assert isinstance(task.transport, AsyncHttpStep)

    task.execute()

    rows = [row.asDict() for row in task.output.df.collect()]
    all_data = [row["url"] for row in rows]

    # Assert the responses_urls
    assert len(all_data) == 2
    assert all_data == [f"{ASYNC_BASE_URL}/get"] * 2


@responses.activate
def test_rest_api_reader(mock_aiohttp):
    """
    Testing the AsyncHttpStep class.
    """

    def request_callback(request):
        import json

        body = [
            {
                "headers": dict(request.headers),
                "url": str(ASYNC_GET_ENDPOINT),
            }
        ]
        return (200, request.headers, json.dumps(body))

    responses.add_callback(
        responses.GET,
        str(ASYNC_GET_ENDPOINT),
        callback=request_callback,
        content_type="application/json",
    )

    transport = HttpGetStep(
        url=str(ASYNC_GET_ENDPOINT),
        headers={"Content-Type": "application/json", "X-Type": "Koheesio RestApiReader Test"},
    )

    spark_schema = StructType(
        [
            StructField(
                "headers",
                StructType(
                    [
                        StructField("Accept", StringType(), True),
                        StructField("Accept-Encoding", StringType(), True),
                        StructField("Connection", StringType(), True),
                        StructField("Content-Type", StringType(), True),
                        StructField("User-Agent", StringType(), True),
                        StructField("X-Type", StringType(), True),
                    ]
                ),
                True,
            ),
            StructField("url", StringType(), True),
        ]
    )

    task = RestApiReader(transport=transport, spark_schema=spark_schema)

    assert isinstance(task.transport, HttpGetStep)

    task.execute()

    rows = [row.asDict() for row in task.output.df.collect()]
    all_data = [{row["url"]: row.get("headers", {}).asDict()["X-Type"]} for row in rows]

    # Assert the responses_urls
    assert len(all_data) == 1
    assert all_data == [{f"{ASYNC_BASE_URL}/get": "Koheesio RestApiReader Test"}]



"""
Transformation to wrap around the pyspark getItem function
"""

from typing import Union

from pyspark.sql import Column

from koheesio.models import Field
from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils import SparkDatatype


def get_item(column: Column, key: Union[str, int]) -> Column:
    """
    Wrapper around pyspark.sql.functions.getItem

    Parameters
    ----------
    column : Column
        The column to get the item from
    key : Union[str, int]
        The key (or index) to get from the list or map. If the column is a list (ArrayType), this should be an integer.
        If the column is a dict (MapType), this should be a string.

    Returns
    -------
    Column
        The column with the item
    """
    return column.getItem(key)


class GetItem(ColumnsTransformationWithTarget):
    """
    Get item from list or map (dictionary)

    Wrapper around `pyspark.sql.functions.getItem`

    `GetItem` is strict about the data type of the column. If the column is not a list or a map, an error will be
    raised.

    Note
    ----
    Only MapType and ArrayType are supported.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to get the item from. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    key : Union[int, str]
        The key (or index) to get from the list or map. If the column is a list (ArrayType), this should be an integer.
        If the column is a dict (MapType), this should be a string. Alias: index

    Example
    -------

    ### Example with list (ArrayType)

    By specifying an integer for the parameter "key", getItem knows to get the element at index n of a list
    (index starts at 0).

    __input_df__:

    | id | content   |
    |----|-----------|
    | 1  | [1, 2, 3] |
    | 2  | [4, 5]    |
    | 3  | [6]       |
    | 4  | []        |

    ```python
    output_df = GetItem(
        column="content",
        index=1,  # get the second element of the list
        target_column="item",
    ).transform(input_df)
    ```

    __output_df__:

    | id | content   | item |
    |----|-----------|------|
    | 1  | [1, 2, 3] | 2    |
    | 2  | [4, 5]    | 5    |
    | 3  | [6]       | null |
    | 4  | []        | null |

    ### Example with a dict (MapType)

    __input_df__:

    | id | content          |
    |----|------------------|
    |  1 | {key1 -> value1} |
    |  2 | {key1 -> value2} |
    |  3 | {key2 -> hello}  |
    |  4 | {key2 -> world}  |

    ```python
    output_df = GetItem(
        column= "content",
        key="key2,
        target_column="item",
    ).transform(input_df)
    ```
    As we request the key to be "key2", the first 2 rows will be null, because
    it does not have "key2".

    __output_df__:

    | id | content          | item  |
    |----|------------------|-------|
    | 1  | {key1 -> value1} | null  |
    | 2  | {key1 -> value2} | null  |
    | 3  | {key2 -> hello}  | hello |
    | 4  | {key2 -> world}  | world |
    """

    class ColumnConfig(ColumnsTransformationWithTarget.ColumnConfig):
        """Limit the data types to ArrayType and MapType."""

        run_for_all_data_type = [SparkDatatype.ARRAY, SparkDatatype.MAP]
        limit_data_type = run_for_all_data_type
        data_type_strict_mode = True

    key: Union[int, str] = Field(
        default=...,
        alias="index",
        description="The key (or index) to get from the list or map. If the column is a list (ArrayType), this should "
        "be an integer. If the column is a dict (MapType), this should be a string. Alias: index",
    )

    def func(self, column: Column) -> Column:
        return get_item(column, self.key)



from pathlib import Path

import pandas as pd

from koheesio.pandas.readers.excel import ExcelReader


def test_excel_reader(data_path):
    # Define the path to the test Excel file and the expected DataFrame
    test_file = Path(data_path) / "readers" / "excel_file" / "dummy.xlsx"
    expected_df = pd.DataFrame(
        {
            "a": ["foo", "so long"],
            "b": ["bar", "and thanks"],
            "c": ["baz", "for all the fish"],
            "d": [None, 42],
            "e": pd.to_datetime(["1/1/24", "1/1/24"], format="%m/%d/%y"),
        }
    )

    # Initialize the ExcelReader with the path to the test file and the correct sheet name
    reader = ExcelReader(path=test_file, sheet_name="sheet_to_select", header=0)

    # Execute the reader
    reader.execute()

    # Assert that the output DataFrame is as expected
    pd.testing.assert_frame_equal(reader.output.df, expected_df, check_dtype=False)



"""
Test the Split transformations
"""

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.split import SplitAll, SplitAtFirstMatch
from koheesio.spark.utils import show_string

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)

string_data = [
    ["Banana lemon orange", 1000, "USA"],
    ["Carrots Blueberries", 1500, "USA"],
    ["Beans", 1600, "USA"],
]
empty_data = [
    ["   ", 1000, "USA"],
    ["", 1500, "USA"],
]
columns = ["product", "amount", "country"]
default_params = dict(
    split_pattern="\\s",
    column="product",
    target_column="split",
)


@pytest.mark.parametrize(
    "input_values,data,schema,expected",
    [
        (
            # description: Test SplitAll
            # input values
            default_params,
            # input data and schema
            string_data,
            columns,
            # expected data
            [
                ["Banana", "lemon", "orange"],
                ["Carrots", "Blueberries"],
                ["Beans"],
            ],
        ),
        (
            # description: Test SplitAll - Empty string split
            # input values
            default_params,
            # input data and schema
            empty_data,
            columns,
            # expected data
            # 3 spaces in the df_empty means 3 patterns to split. Resulting in 4 elements.
            [
                ["", "", "", ""],
                [""],
            ],
        ),
        (
            # description: Test SplitAll - Transform multi character string split
            # input values
            {**default_params, "split_pattern": "an"},
            # input data and schema
            string_data,
            columns,
            # expected data
            [
                ["B", "", "a lemon or", "ge"],
                ["Carrots Blueberries"],
                ["Be", "s"],
            ],
        ),
    ],
)
def test_split_all(input_values, data, schema, expected, spark):
    input_df = spark.createDataFrame(data, schema)

    split_all = SplitAll(**input_values)
    filter_column = split_all.target_column
    output_df = split_all.transform(df=input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{show_string(output_df, 20, 20, False)}")

    actual = [row.asDict()[filter_column] for row in output_df.collect()]
    assert actual == expected


# TODO: not supported yet
#
#     def test_transform_nested_df(self):
#         output_df = (
#             SplitColumn(
#                 df=self.sample_nested_df,
#                 params={
#                     "split_pattern": "l",
#                     "source_column": "payload.col2.col3",
#                 },
#                 target_alias="split",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         output_list_split = output_df.select("split").rdd.map(lambda r: r.split).collect()
#
#         self.assertIn(["va", "ue1"], output_list_split)


@pytest.mark.parametrize(
    "input_values,data,schema,expected",
    [
        (
            # description: Test SplitAtFirstMatch - Check the flow when retrieve_first_part is set to true
            # input values
            {**default_params, "retrieve_first_part": True},
            # input data and schema
            string_data,
            columns,
            # expected data
            ["Banana", "Carrots", "Beans"],
        ),
        (
            # description: Test SplitAtFirstMatch - Check the flow when retrieve_first_part is set to false
            # input values
            {**default_params, "retrieve_first_part": False},
            # input data and schema
            string_data,
            columns,
            # expected data
            ["lemon", "Blueberries", ""],
        ),
        (
            # description: Test SplitAtFirstMatch - Check empty column
            # input values
            default_params,
            # input data and schema
            empty_data,
            columns,
            # expected data
            ["", ""],
        ),
        (
            # description: Test SplitAtFirstMatch - Check multi character split
            # This is the same as what is mentioned in the docstring for the class
            # input values
            {**default_params, "split_pattern": "an"},
            # input data and schema
            string_data,
            columns,
            # expected data
            ["B", "Carrots Blueberries", "Be"],
        ),
    ],
)
def test_split_at_first_match(input_values, data, schema, expected, spark):
    input_df = spark.createDataFrame(data, schema)

    split_at_first_match = SplitAtFirstMatch(**input_values)
    filter_column = split_at_first_match.target_column
    output_df = split_at_first_match.transform(df=input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{show_string(output_df, 20, 20, False)}")

    actual = [row.asDict()[filter_column] for row in output_df.collect()]
    assert actual == expected


# TODO: not supported yet
#
#     def test_transform_nested_df_single_character_split(self):
#         """
#         Check execute method for a nested dataframe
#         """
#         output_df = (
#             SplitColumnAtFirstMatch(
#                 df=self.sample_nested_df,
#                 params={
#                     "split_pattern": "l",
#                     "source_column": "payload.col2.col3",
#                 },
#                 target_alias="payload.col2.col3_split",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         json_output = output_df.toJSON().map(lambda j: json.loads(j)).collect()
#
#         actual_col3_split1 = json_output[0].get("payload").get("col2").get("col3_split")
#
#         self.assertEqual(actual_col3_split1, "va")
#
#         output_df = (
#             SplitColumnAtFirstMatch(
#                 df=self.sample_nested_df,
#                 params={
#                     "split_pattern": "l",
#                     "source_column": "payload.col2.col3",
#                     "retrieve_first_part": False,
#                 },
#                 target_alias="payload.col2.col3_split",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#         json_output = output_df.toJSON().map(lambda j: json.loads(j)).collect()
#
#         actual_col3_split2 = json_output[0].get("payload").get("col2").get("col3_split")
#
#         self.assertEqual(actual_col3_split2, "ue1")



import pytest

from koheesio.spark.readers.dummy import DummyReader

pytestmark = pytest.mark.spark


def test_reader(spark):
    test_reader = DummyReader(range=1)
    actual = test_reader.read().head().asDict()
    expected = {"id": 0}
    assert actual == expected


def test_execute(spark):
    test_reader = DummyReader(range=1)
    actual = test_reader.execute().df.head().asDict()
    expected = {"id": 0}
    assert actual == expected



"""A type-safe Python framework for building efficient data pipelines

Koheesio is a Python framework for building type-safe and efficient data pipelines, particularly useful for
large-scale data processing. It promotes modularity, composability, and collaboration, allowing for the creation of
complex pipelines from simple, reusable components.

Leveraging Pydantic, it ensures predictable pipeline execution and structured configurations within pipeline components.

With its simple and straightforward API, Koheesio makes it easy for developers to build and manage data pipelines,
enhancing productivity and code maintainability.
"""

LICENSE_INFO = "Licensed as Apache 2.0"
SOURCE = "https://github.com/Nike-Inc/koheesio"

__version__ = "0.10.6"
__logo__ = (
    75,
    (
        b"\x1f\x8b\x08\x00TiGf\x02\xff}\x91\xbb\r\xc30\x0cD{Nq\x1bh\n\x01\x16R\xa4pK@\x8bh\xf8\xe8\xf8\x89\xe9\x04\xf0\x15"
        b"\xc4\x91\x10\x9f(J`z\xbd4B\xea8J\xf2\xa01T\x02\x01,\x0b\x85Q\x92\x07\xe9\x9cK\x92\xd1,\xe0mRBL\x9c\xa6\x9b\xee"
        b"\xeet)\x07Av\xc9/\x0b\x98\x93\xb4=\xd1v\xa4\xf5NG\xc6\xe5\xce\x93nk\x8d\x81\xf5\xed\x92\x80AmC\xbb\xde,.\x7f\x1fc"
        b"\x0fU\xa79\x19\x82\x16]\x1248\x8f\xa5\x7f\x1c|\x92\xe2\xb8\xa59\xfd\xa5\x86\x8b.I\x9a\xf3\xd4W\x80\x8a\xd3\x9e"
        b"\xfb\xba\\\xecm\x9f#\xee\xea\x92}M+\xffb\xb7\xb2\xc4\xc4K\x88Zui\xda\xedD\xfb\x00\xcfU6\xd3_\x02\x00\x00"
    ),
)
__short_description__ = __doc__.split("\n", maxsplit=1)[0]
__about__ = f"""Koheesio - v{__version__}
{__short_description__}
{LICENSE_INFO}
Source: {SOURCE}
"""


# fmt: off
def _about() -> str:  # pragma: no cover
    """Return the Koheesio logo and version/about information as a string
    Note: this code is not meant to be readable, instead it is written to be as compact as possible
    """
    import gzip
    _w, _logo = __logo__
    _h, _v, _n = "\u2500", "\u2502", "\n"
    return gzip.decompress(_logo).decode() + (
        _n + '\u256d' + _h * _w + "\u256e" + _n +
        f"{_n.join([f'{_v}{_l:^{_w}}{_v}' for _l in __about__.split(_n)[:-1]])}" +
        _n + "\u2570" + _h * _w + "\u256f"
    )
# fmt: on


if __name__ == "__main__":
    print(_about())
